# -*- coding: utf-8 -*-

"""

  HTTP semantics
  ~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# session API
from .. import session

# core
from canteen.base import logic
from canteen.core import runtime

# canteen util
from canteen.util import config
from canteen.util import decorators


def url(name_or_route, route=None, **kwargs):

  """ Decorator for binding a URL (potentially with a name) to some sort of
      handler, which can really be anything from raw types like a ``str`` or a
      ``tuple`` to a full handler/function/class to a foreign WSGI callable.

      :param name_or_route: Either the route itself (if no name is to be
        provided) or a string name for the route, which can be used to easily
        build URL's to named pages which do not rely on the underlying URL
        structure. Type is always ``basestring``.

      :param route: If a name is provided in ``name_or_route``, this should be
        a ``basestring`` route mapping in Werkzeug style.

      :param kwargs: Keyword arguments to pass to ``add_route``, which is the
        method used to associate the target handler with the given URL. One
        ``kwarg`` is used locally - like many decorators in Canteen, you can
        pass another callable as ``wrap=callable`` and it will be used to chain
        decoration before construction.

      :returns: Decorated target, unchanged, after registration. """

  # extract route and name (optional, but first arg if specified)
  name, route = (None, name_or_route) if not route else (name_or_route, route)

  # must provide at least a route
  if not route: raise ValueError('Cannot bind to a URL with an empty route.')

  # inject the rule factory
  def bind_route(target):

    """ Wrapped closure that applies a URL binding to a target to-be-decorated.

        :param target: Handler to be decorated with this URL binding, by
          associating the two in the currently-active runtime.

        :returns: Decorated ``target``, unchanged, after registration. """

    wrap = kwargs.get('wrap') or (lambda x: x)  # check wrap
    if 'wrap' in kwargs: del kwargs['wrap']

    for entry in ((route,) if not isinstance(route, tuple) else route):
      HTTPSemantics.add_route((entry, name), target, **kwargs)
    return wrap(target)

  return bind_route


with runtime.Library('werkzeug', strict=True) as (library, werkzeug):

  # werkzeug internals
  wsgi, utils, routing, wrappers, exceptions = library.load(
    'wsgi',
    'utils',
    'routing',
    'wrappers',
    'exceptions')


  @decorators.bind('http', namespace=False)
  class HTTPSemantics(logic.Logic):

    """ Packages and provides basic logic to support communication via Hypertext
        Transfer Protocol, versions 1.0-1.1 (support for HTTPv2 soon). """

    # @TODO(sgammon): rename __aliases__ as it clashes with DI
    __aliases__, __map__, __router__ = (
      {}, None, None)  # aliases to routes, routing map cache, route cache


    # == Base Classes == #
    HTTPException = exceptions.HTTPException


    class HTTPRequest(wrappers.BaseRequest,
                      wrappers.AcceptMixin,
                      wrappers.UserAgentMixin,
                      wrappers.AuthorizationMixin,
                      wrappers.CommonRequestDescriptorsMixin):

      """ Base HTTP request wrapper object, backed by Werkzeug's ``BaseRequest``
          and rich functionality mixins.

          Provides the following mixin functionality:
          - ``AcceptMixin`` - HTTP accept header handling
          - ``UserAgentMixin`` - HTTP user-agent header handling
          - ``AuthorizationMixin`` - HTTP authorization header handling
          - ``CommonRequestDescriptorsMixin`` - extra utility attributes """

      # internal slot & accesstor for session
      __session__, session = None, property(lambda self: self.__session__)

      def set_session(self, session_obj, engine=None):

        """ Set the current session by force.

            :param session_obj: Session object to force-assign as the current
              user's session. Either a ``dict`` or a manually-constructed
              :py:class:`canteen.logic.session.Session` object.

            :param engine: Engine to use for committing the session. Defaults
              to ``None``, in which case the currently-active ``engine`` is
              used automatically.

            :returns: ``self``, for chainability. """

        if session_obj and not isinstance(session_obj, session.Session):
          if 'uuid' not in session_obj and 'id' not in session_obj:
            session_obj = None  # no ID or UUID == no session
          else:  # pragma: no cover
            session_obj = session.Session.load((
              session_obj['uuid' if 'uuid' in session_obj else 'id']),
              data=session_obj)

        if self.__session__:  # pragma: no cover
          existing_session, existing_engine = self.__session__
          if not engine: engine = existing_engine

        return setattr(self, '__session__', (session_obj, engine)) or (
          self.__session__)


    class HTTPResponse(wrappers.BaseResponse,
                       wrappers.ETagResponseMixin,
                       wrappers.ResponseStreamMixin,
                       wrappers.WWWAuthenticateMixin,
                       wrappers.CommonResponseDescriptorsMixin):

      """ Base HTTP response wrapper object, backed by Werkzeug's
          ``BaseResponse`` and rich functionality mixins.

          Provides the following mixin functionality:
          - ``ETagResponseMixin`` - provides ETag header handling
          - ``ResponseStreamMixin`` - provides streaming-style response output
          - ``WWWAuthenticateMixin`` - provides HTTP authentication response
          - ``CommonResponseDescriptorsMixin`` - provides common attributes """


    #### ==== Internals ==== ####

    # noinspection PyMethodParameters
    @decorators.classproperty
    def config(cls):  # pragma: no cover

      """ Class-level property accessor for framework/app configuration.

          :returns: Currently-active instance of
            :py:class:`canteen.util.config.Config`. """

      return config.Config().get('http', {'debug': True})

    @property
    def base_headers(self):

      """ Prepare a set of default (*base*) HTTP response headers to be included
          by-default on any HTTP response.

          :returns: ``list`` of ``tuples``, where each is a pair of ``key``-bound
            ``value`` mappings. Because HTTP headers can be repeated, a ``dict``
            is not usable in this instance. """

      return filter(lambda x: x and x[1], [
        ('Cache-Control', 'no-cache; no-store')])

    #### ==== Routing ==== ####
    @classmethod
    def add_route(cls, route, target, **kwargs):

      """ Bind a handler (``target``) to a URL (``route``) and let it respond
          to later requests that match that ``route``.

          :param route: Route to bind the ``target`` to. This is a Werkzeug-
            style URL specification string.

          :param target: Handler that should be dispatched for requests that
            match the given ``route``.

          :param kwargs: Keyword arguments to pass along to the ``target`` upon
            dispatching URLs against it.

          :returns: ``cls``, for chainability. """

      route, name = route

      # compile route and set in routing cache
      cls.router.set(route, (name, target, kwargs))
      if name: cls.__aliases__[name] = route
      return cls

    @classmethod
    def resolve_route(cls, name):

      """ Resolve a handler for a given route ``name``.

          :param name: Name to find a handler for in the currently-active
            ``cls.router``.

          :returns: ``handler`` bound to name ``name`` via a route, or ``None``
            if no such handler could be found. """

      if name in cls.__aliases__:
        name, handler, route_args = cls.router.get(cls.__aliases__[name])
        return handler
      return None

    @classmethod
    def new_request(cls, environ):

      """ Utility to spawn a new HTTP request wrapper object. Constructs an
          instance of :py:class:`cls.HTTPRequest`, which is usually backed by
          Werkzeug's HTTP internals.

          :param environ: WSGI environment (``dict``) to create the request
            wrapper from.

          :returns: Instance of ``cls.HTTPRequest``, filled out with
            ``environ``. """

      return cls.HTTPRequest(environ)

    @classmethod
    def new_response(cls, *args, **kwargs):

      """ Utility to spawn a new HTTP response wrapper obejct. Constructs an
          instance of :py:class:`cls.HTTPResponse`, which is usually backed by
          Werkzeug's HTTP internals.

          :param args: Positional arguments to pass to the HTTP response wrapper
            object's constructor.

          :param kwargs: Keyword arguments to pass to the HTTP response wrapper
            object's constructor.

          :returns: Instance of ``cls.HTTPResponse``, with arguments passed
            along verbatim. """

      return cls.HTTPResponse(*args, **kwargs)

    @decorators.classproperty
    def router(cls):

      """ Class-level property accessor for the current URL router, which keeps
          track of associations between URLs and handlers, which respond to
          requests for a set of URLs.

          Accessing this attribute the first time spawns a new cache for the
          router, and subsequent accesses re-use that cache.

          :returns: Instance of :py:class:`canteen.logic.cache.Cache` to use for
            resolving handlers. """

      from canteen.logic import cache

      if not cls.__router__:
        setattr(cls, '__router__', cache.Caching.spawn('router'))
      return cls.__router__

    @decorators.classproperty
    def routes(cls):

      """ Class-level property generator for the currently-active set of URL
          routes. Makes use of the local ``cls.router`` to iterate over all
          known routes.

          :returns: Yields ``cls.router``-bound routes, one-at-a-time. """

      for _url, (name, target, kwargs) in cls.router.items():
        yield routing.Rule(_url, endpoint=name, **kwargs)

    @decorators.classproperty
    def route_map(cls):

      """ Class-level property accessor for the current routing map, which is
          an instance of :py:class:`werkzeug.routing.Map`, composed of the
          currently-active ruleset for URL routing.

          :returns: Instance of :py:class:`werkzeug.routing.Map`. """

      if not cls.__map__:
        setattr(cls, '__map__', routing.Map([route for route in cls.routes]))
      return cls.__map__

    #### ==== Utilities ==== ####
    @decorators.bind('error', wrap=classmethod)
    def error(cls, code):

      """ Force an HTTP error by integer ``code``, like ``404`` (for error
          state "Not Found") or ``500``. Resolves the appropriate internal
          exception to raise and raises it immeditely.

          :param code: ``int`` HTTP error state code to raise.

          :raises werkzeug.exceptions.HTTPException: Resolves proper exception
            class to raise, which is always a subclass of ``HTTPException``,
            which is provided by Werkzeug. """

      exceptions.abort(code)

    @decorators.bind('redirect', wrap=classmethod)
    def redirect(cls, url=None, name=None, permanent=False, code=None,
                 *args, **kwargs):

      """ Prepare an HTTP redirect to ``url`` or ``name``, where ``name`` is a
          named URL bound earlier via the ``url`` decorator.

          :param url: URL (absolute or relative) to redirect to.

          :param name: Named endpoint (bound previously via ``url`` decorator)
            to generate a URL for and redirect to.

          :param permanent: ``bool`` flag deciding whether to issue a
            'permanent' redirect, which uses the HTTP code 301 instead of the
             standard "302 Found". Defaults to ``False``, in which case an
             HTTP 302 is issued.

          :param code: ``int`` HTTP code to issue for the redirect. Must be
            greater than 299 and less than 399, indicating it is part of the
            standard range of 'redirect' codes in the HTTP spec. Used over the
            ``permanent`` flag if both are issued.

          :returns: Prepared HTTP redirect, ready to be returned from a handler
            dispatch flow. """

      # sanity checks
      if (not url and not name) or (url and name):  # pragma: no cover
        raise TypeError('Must provide either a URL or name to redirect to -'
                        ' not both and at least one or the other.'
                        ' Got "%s" and "%s".' % (url, name))
      return utils.redirect(url if url else cls.url_for(name, *args, **kwargs),
                            code=302 if not permanent else 301)

    # @TODO(sgammon): is this necessary?
    @decorators.bind('response', wrap=classmethod)
    def response(cls, *args, **kwargs):  # pragma: no cover

      """ Property accessor to generate a ``wrappers.Response`` object on the
          fly via DI.

          :param args: Positional arguments to pass to the response wrapper
            constructor.

          :param kwargs: Keyword arguments to pass to the response wrapper
            constructor. """

      return wrappers.Response(*args, **kwargs)

    #### ==== HTTP Methods ==== ####

    # default to a `Method Not Allowed`
    GET = decorators.bind('GET')(lambda self: self.error(405))
    POST = decorators.bind('POST')(lambda self: self.error(405))
    PUT = decorators.bind('PUT')(lambda self: self.error(405))
    HEAD = decorators.bind('HEAD')(lambda self: self.error(405))
    OPTIONS = decorators.bind('OPTIONS')(lambda self: self.error(405))
    TRACE = decorators.bind('TRACE')(lambda self: self.error(405))



# -*- coding: utf-8 -*-

"""

  model tests
  ~~~~~~~~~~~

  tests canteen's model classes.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import abc
import json
import inspect

# appconfig
try:
  import config; _APPCONFIG = True
except ImportError as e:  # pragma: no cover
  _APPCONFIG = False
else:  # pragma: no cover
  # set debug mode for all model-related stuff
  class Config(object):
    config = {}
    debug = True
  config = Config()
  config.config = {}

  if 'canteen.model' not in config.config:
    config.config['canteen.model'] = {}
  config.config['canteen.model']['debug'] = True
  for k in filter(lambda x: x.startswith('canteen.model'),
                              config.config.iterkeys()):
    config.config[k]['debug'] = True

# canteen model API
from canteen import model, core
from canteen.model import adapter
from canteen.model import exceptions

# canteen tests
from canteen.test import FrameworkTest

# canteen utils
from canteen.util import struct as datastructures


## == Test Models == ##

class TestCar(model.Model):

  """ An automobile. """

  make = basestring, {'indexed': True}
  model = basestring, {'indexed': True}
  year = int, {'choices': xrange(1900, 2015)}
  color = basestring, {'choices': (
                        'blue', 'green', 'red', 'silver', 'white', 'black')}


class TestPerson(model.Vertex):

  """ A human being. """

  firstname = basestring, {'indexed': True}
  lastname = basestring, {'indexed': True}
  active = bool, {'default': True}
  cars = TestCar, {'repeated': True}
  cars_ref = model.Key, {'repeated': True}
  cars_embedded = TestCar, {'repeated': True, 'embedded': True}


class Friendship(TestPerson > TestPerson):

  """ A friendship between people. """

  year_met = int


## ModelTests
class ModelTests(FrameworkTest):

  """ Tests `model.Model` and `model.AbstractModel`. """

  def test_construct_model(self):

    """ Test constructing a `Model` manually """

    # construct our car record
    car = TestCar(make='BMW', model='M3', year=2013, color='white')

    # construct our person record
    person = TestPerson()
    person.firstname = 'John'
    person.cars = [car]

    # perform tests
    self.assertIsInstance(car, TestCar)
    self.assertIsInstance(person, TestPerson)
    self.assertEqual(person.firstname, 'John')
    self.assertIsInstance(person.cars, list)
    self.assertEqual(len(person.cars), 1)
    self.assertIsInstance(person.cars[0], TestCar)

    # test defaults
    self.assertEqual(person.active, True)

    # test unsets
    self.assertEqual(person.lastname, None)

    # test __repr__
    cls = str(TestPerson)
    obj = str(person)

    # test class representations
    self.assertTrue(("Person" in cls))
    self.assertTrue(("lastname" in cls))
    self.assertTrue(("firstname" in cls))

    # test object representations
    self.assertTrue(("Person" in obj))
    self.assertTrue(("lastname" in obj))
    self.assertTrue(("firstname" in obj))
    self.assertTrue(("John" in obj))

  def test_invalid_model_adapter(self):

    """ Test using invalid model adapter, which should raise `RuntimeError` """

    with self.assertRaises(RuntimeError):

      ## InvalidAdapterModel
      # Tests an invalid, but explicitly listed, model adapter.
      class InvalidAdapterModel(model.Model):

        __adapter__ = 'DumbInvalidModelAdapter'

        prop1 = basestring

  def test_model_inheritance(self):

    """ Test proper inheritance structure for `Model` """

    self.assertTrue(issubclass(TestCar, model.Model))
    self.assertTrue(issubclass(TestPerson, model.Model))
    self.assertTrue(issubclass(model.Model, model.AbstractModel))

  def test_model_schema(self):

    """ Test that there's a proper schema spec on `Model` """

    # check lookup
    self.assertTrue(hasattr(TestPerson, '__lookup__'))
    self.assertIsInstance(TestPerson.__lookup__, frozenset)

    # check property descriptors
    self.assertTrue(hasattr(TestPerson, 'firstname'))
    self.assertTrue(hasattr(TestPerson, 'lastname'))
    self.assertTrue(hasattr(TestPerson, 'cars'))

    # check kind
    self.assertTrue(hasattr(TestPerson, 'kind'))
    self.assertIsInstance(TestPerson.kind(), basestring)

    # check set/get
    self.assertTrue(hasattr(TestPerson, '_get_value'))
    self.assertTrue(hasattr(TestPerson, '_set_value'))
    self.assertTrue(inspect.ismethod(TestPerson._get_value))
    self.assertTrue(inspect.ismethod(TestPerson._set_value))

    # check key
    self.assertTrue(hasattr(TestPerson, 'key'))
    self.assertTrue(hasattr(TestPerson(), '__key__'))

    # should not have key until instantiation
    self.assertTrue((not hasattr(TestPerson, '__key__')))

  def test_model_set_attribute(self):

    """ Test setting an unknown and known attribute on `Model` """

    # try construction assignment
    john = TestPerson(firstname='John')
    self.assertEqual(john.firstname, 'John')

    # re-assign
    john.firstname = 'Blabs'
    self.assertEqual(john.firstname, 'Blabs')

    # try assigning missing property
    with self.assertRaises(AttributeError):
      john.blabs = 'John'

  def test_model_adapter(self):

    """ Test that adapter is attached correctly to `Model` """

    # make sure it's on the classlevel
    self.assertTrue(hasattr(TestPerson, '__adapter__'))
    self.assertIsInstance(TestPerson.__adapter__, adapter.ModelAdapter)

  def test_model_stringify(self):

    """ Test the string representation of a `Model` object """

    self.assertIsInstance(TestPerson().__repr__(), basestring)

  def test_model_kind(self):

    """ Test that `Model.kind` is properly set """

    # test class-level kind
    self.assertIsInstance(TestPerson.kind(), basestring)
    self.assertEqual(TestPerson.kind(), "TestPerson")

    # test object-level kind
    john = TestPerson()
    self.assertIsInstance(john.kind(), basestring)
    self.assertEqual(john.kind(), "TestPerson")

  def test_abstract_model(self):

    """ Test that `AbstractModel` works abstractly """

    # make sure it's ABC-enabled
    self.assertTrue((not isinstance(model.Model, abc.ABCMeta)))

    # try directly-instantiation
    with self.assertRaises(exceptions.AbstractConstructionFailure):
      (model.AbstractModel())

  def test_concrete_model(self):

    """ Test that `Model` works concretely """

    ## test simple construction
    self.assertIsInstance(TestPerson(), TestPerson)

    ## SampleModel
    # Test parent model class.
    class SampleModel(model.Model):

      """ Test parent model class. """

      parent = basestring

    ## SampleSubModel
    # Test child model class.
    class SampleSubModel(SampleModel):

      """ Test child model class. """

      child = basestring

    ## test properties
    self.assertTrue(hasattr(SampleModel, 'parent'))
    self.assertTrue((not hasattr(SampleModel, 'child')))

    ## test submodel properties
    self.assertTrue(hasattr(SampleSubModel, 'child'))
    self.assertTrue(hasattr(SampleSubModel, 'parent'))

    ## test recursive subclassing
    self.assertIsInstance(SampleModel(), model.Model)
    self.assertIsInstance(SampleSubModel(), SampleModel)
    self.assertIsInstance(SampleSubModel(), model.Model)

  def test_model_to_dict(self, method='to_dict'):

    """ Test flattening a `Model` into a raw dictionary """

    # sample person
    p = TestPerson(firstname='John')
    raw_dict = getattr(p, method)()

    if method == 'to_dict':
      # try regular to_dict
      self.assertEqual(len(raw_dict), 2)
      self.assertIsInstance(raw_dict, dict)
      self.assertEqual(raw_dict['firstname'], 'John')  # we set this explicitly

      # this is defaulted, should export
      self.assertEqual(raw_dict['active'], True)

      with self.assertRaises(KeyError):
        raw_dict['lastname']
    return raw_dict

  def test_model_update_with_dict(self):

    """ Test updating a `Model` from a `dict` """

    # sample person
    p = TestPerson(firstname='John')
    update = {
      'firstname': 'Sup',
      'lastname': 'Bleebs'
    }

    p.update(update)

    assert p.firstname == 'Sup'
    assert p.lastname == 'Bleebs'

  def test_model_to_dict_schema(self):

    """ Test flattening a `Model` class into a schema dictionary """

    schema = TestPerson.to_dict_schema()
    assert 'firstname' in schema
    assert isinstance(schema['firstname'], model.Property)
    assert schema['firstname']._basetype == basestring

  def test_model_to_dict_all_arguments(self, method='to_dict'):

    """ Test using `Model.to_dict` with the `all` flag """

    # sample person
    p = TestPerson(firstname='John')
    all_dict = getattr(p, method)(_all=True)

    if method == 'to_dict':
      # test dict with `all`
      self.assertEqual(len(all_dict), len(p.__lookup__))
      self.assertEqual(all_dict['firstname'], 'John')
      self.assertEqual(all_dict['lastname'], None)
      self.assertEqual(all_dict['active'], True)
    return all_dict

  def test_model_to_dict_with_filter(self, method='to_dict'):

    """ Test using `Model.to_dict` with a filter function """

    # sample person
    p = TestPerson(firstname='John')

    # should filter out 'active'
    filtered_dict = getattr(p, method)(filter=lambda x: len(x[0]) > 7)

    if method == 'to_dict':
      # test filter
      self.assertEqual(len(filtered_dict), 1)
      self.assertIsInstance(filtered_dict, dict)
      self.assertEqual(filtered_dict['firstname'], 'John')

      with self.assertRaises(KeyError):
        filtered_dict['active']
    return filtered_dict

  def test_model_to_dict_with_include(self, method='to_dict'):

    """ Test using `Model.to_dict` with an inclusion list """

    # sample person
    p = TestPerson(firstname='John')
    included_dict = getattr(p, method)(include=('firstname', 'lastname'))

    if method == 'to_dict':
      # should still include `lastname` as 'None'
      self.assertEqual(len(included_dict), 2)
      self.assertIsInstance(included_dict, dict)
      self.assertEqual(included_dict['firstname'], 'John')
      self.assertEqual(included_dict['lastname'], None)

      with self.assertRaises(KeyError):
        included_dict['active']  # should not include `active`
    return included_dict

  def test_model_to_dict_with_exclude(self, method='to_dict'):

    """ Test using `Model.to_dict` with an exclusion list """

    # sample person
    p = TestPerson(firstname='John')
    excluded_dict = getattr(p, method)(exclude=('active',))

    if method == 'to_dict':
      # test exclude
      self.assertEqual(len(excluded_dict), 1)
      self.assertIsInstance(excluded_dict, dict)
      self.assertEqual(excluded_dict['firstname'], 'John')

      with self.assertRaises(KeyError):
        excluded_dict['active']  # should not include `active`
    return excluded_dict

  def test_model_to_dict_with_map(self, method='to_dict'):

    """ Test using `Model.to_dict` with a map function """

    # sample person
    p = TestPerson(firstname='John')
    mapped_dict = getattr(p, method)(
                             map=lambda x: tuple([x[0] + '-cool', x[1]]))

    if method == 'to_dict':
      # test map
      self.assertEqual(len(mapped_dict), 2)
      self.assertIsInstance(mapped_dict, dict)
      self.assertEqual(mapped_dict['firstname-cool'], 'John')
      self.assertEqual(mapped_dict['active-cool'], True)
    return mapped_dict

  def test_model_to_dict_convert_keys(self):

    """ Test flattening a `Model` instance with key references to a dict """

    # sample person
    p = TestPerson(
      key=model.Key(TestPerson, 'john'), firstname='John', lastname='Doe')

    # john's cars
    bmw = TestCar(
      key=model.Key(TestCar, 'bmw'),
      make='BMW', model='M3', color='white', year=1998)
    civic = TestCar(
      key=model.Key(TestCar, 'civic'),
      make='Honda', model='Civic', color='black', year=2001)

    p.cars_ref = (bmw.key, civic.key)

    _john_raw = p.to_dict(convert_keys=False)  # without convert keys
    _john_converted = p.to_dict(convert_keys=True)  # with convert_keys

    assert bmw.key in _john_raw['cars_ref'], (
      "expected key for car `BMW` but found none"
      " in object: '%s'." % _john_raw)

    assert civic.key in _john_raw['cars_ref'], (
      "expected key for car `civic` but found none"
      " in object: '%s'." % _john_raw)

    assert bmw.key.urlsafe() in _john_converted['cars_ref'], (
      "expected converted key for car `BMW` but found none"
      " in object: '%s'." % _john_converted)

    assert civic.key.urlsafe() in _john_converted['cars_ref'], (
      "expected converted key for car `civic` but found none"
      " in object: '%s'." % _john_converted)

  def test_model_to_dict_convert_models(self):

    """ Test flattening a `Model` instance with submodels to a dict """

    # sample person
    p = TestPerson(
      key=model.Key(TestPerson, 'john'), firstname='John', lastname='Doe')

    # john's cars
    bmw = TestCar(
      key=model.Key(TestCar, 'bmw'),
      make='BMW', model='M3', color='white', year=1998)
    civic = TestCar(
      key=model.Key(TestCar, 'civic'),
      make='Honda', model='Civic', color='black', year=2001)

    p.cars_ref = (bmw.key, civic.key)
    p.cars_embedded = (bmw, civic)

    _john_raw = p.to_dict(convert_keys=False, convert_models=False)
    _john_converted = p.to_dict(convert_keys=True, convert_models=True)

    assert bmw in _john_raw['cars_embedded'], (
      "expected object for car `BMW` but found none"
      " in object: '%s'." % _john_raw)

    assert civic in _john_raw['cars_embedded'], (
      "expected object for car `civic` but found none"
      " in object: '%s'." % _john_raw)

    assert bmw.key in _john_raw['cars_ref'], (
      "expected key for car `BMW` but found none"
      " in object: '%s'." % _john_raw)

    assert civic.key in _john_raw['cars_ref'], (
      "expected key for car `civic` but found none"
      " in object: '%s'." % _john_raw)

    assert bmw.key.urlsafe() in _john_converted['cars_ref'], (
      "expected converted key for car `BMW` but found none"
      " in object: '%s'." % _john_converted)

    assert civic.key.urlsafe() in _john_converted['cars_ref'], (
      "expected converted key for car `civic` but found none"
      " in object: '%s'." % _john_converted)

    assert len(_john_converted['cars_embedded']) == 2
    for i in _john_converted['cars_embedded']:
      assert isinstance(i, dict), (
        "submodels are expected to convert to dicts"
        " when `convert_models` is active")

  def test_JSON_model_format(self):

    """ Test serializing a `Model` into a JSON struct """

    # sample person
    p = TestPerson(firstname='John', lastname='Doe')

    # prepare mini testsuite
    def test_json_flow(original, js=None):

      if not js:
        # execute for the caller
        original, js = original('to_dict'), original('to_json')

      # test string
      self.assertTrue(len(js) > 0)
      self.assertIsInstance(js, basestring)

      # test decode
      decoded = json.loads(js)
      self.assertIsInstance(decoded, dict)
      self.assertEqual(len(original), len(decoded))

      # test property values
      for key in original:
        self.assertEqual(original[key], decoded[key])

    # test regular to_json
    test_json_flow(p.to_dict(), p.to_json())

    # test all to_dict permutations with json
    test_structs = {
      'raw_dict': self.test_model_to_dict,
      'all_dict': self.test_model_to_dict_all_arguments,
      'mapped_dict': self.test_model_to_dict_with_map,
      'filtered_dict': self.test_model_to_dict_with_filter,
      'included_dict': self.test_model_to_dict_with_include,
      'excluded_dict': self.test_model_to_dict_with_exclude
    }

    # test each dict => json flow
    test_json_flow(test_structs['raw_dict'])
    test_json_flow(test_structs['all_dict'])
    test_json_flow(test_structs['mapped_dict'])
    test_json_flow(test_structs['filtered_dict'])
    test_json_flow(test_structs['included_dict'])
    test_json_flow(test_structs['excluded_dict'])

  def test_inflate_model_from_json(self):

    """ Test inflating a `Model` object from a JSON string """

    obj = {'firstname': 'John', 'lastname': 'Doe'}
    json_string = json.dumps(obj)

    # load into object
    p = TestPerson.from_json(json_string)
    assert p.firstname == 'John'
    assert p.lastname == 'Doe'

  with core.Library('msgpack') as (library, msgpack):

    import msgpack  # force re-import


    def test_msgpack_model_format(self):

      """ Test serializing a `Model` into a msgpack struct """

      # sample person
      p = TestPerson(firstname='John', lastname='Doe')

      # prepare mini testsuite
      def test_msgpack_flow(original, mp=None):

        if not mp:
          # execute for the caller
          original, mp = original('to_dict'), original('to_msgpack')

        # test string
        self.assertTrue(len(mp) > 0)
        self.assertIsInstance(mp, basestring)

        # test decode
        decoded = self.msgpack.unpackb(mp)
        self.assertIsInstance(decoded, dict)
        self.assertEqual(len(original), len(decoded))

        # test property values
        for key in original:
          self.assertEqual(original[key], decoded[key])

      # test regular to_msgpack
      test_msgpack_flow(p.to_dict(), p.to_msgpack())

      # test all to_dict permutations with msgpack
      test_structs = {
        'raw_dict': self.test_model_to_dict,
        'all_dict': self.test_model_to_dict_all_arguments,
        'mapped_dict': self.test_model_to_dict_with_map,
        'filtered_dict': self.test_model_to_dict_with_filter,
        'included_dict': self.test_model_to_dict_with_include,
        'excluded_dict': self.test_model_to_dict_with_exclude
      }

      # test each dict => msgpack flow
      test_msgpack_flow(test_structs['raw_dict'])
      test_msgpack_flow(test_structs['all_dict'])
      test_msgpack_flow(test_structs['mapped_dict'])
      test_msgpack_flow(test_structs['filtered_dict'])
      test_msgpack_flow(test_structs['included_dict'])
      test_msgpack_flow(test_structs['excluded_dict'])

    def test_inflate_model_from_msgpack(self):

      """ Test inflating a `Model` object from a msgpack payload """

      obj = {'firstname': 'John', 'lastname': 'Doe'}
      mpack = self.msgpack.dumps(obj)

      # load into object
      p = TestPerson.from_msgpack(mpack)
      assert p.firstname == 'John'
      assert p.lastname == 'Doe'

  def test_explicit(self):

    """ Test a `Model`'s behavior in `explicit` mode """

    # sample people
    s = TestPerson(firstname='Sam')
    p = TestPerson(firstname='John')

    # go into explicit mode
    self.assertEqual(p.__explicit__, False)
    explicit_firstname, explicit_lastname, explicit_active = None, None, None
    with p:
      explicit_firstname, explicit_lastname, explicit_active = (
          p.firstname, p.lastname, p.active)

      # only instances switch modes
      self.assertNotEqual(p.__explicit__, s.__explicit__)

      self.assertEqual(s.lastname, None)
      self.assertEqual(s.active, True)
      self.assertEqual(s.firstname, 'Sam')
      self.assertEqual(p.__explicit__, True)
    self.assertEqual(p.__explicit__, False)

    # test explicit values
    self.assertEqual(explicit_firstname, 'John')

    # default values are not returned in `explicit` mode
    self.assertEqual(explicit_active, datastructures.EMPTY)

    # unset properties are returned as EMPTY in `explicit` mode
    self.assertEqual(explicit_lastname, datastructures.EMPTY)

    # test implicit values
    self.assertEqual(p.firstname, 'John')
    self.assertEqual(p.lastname, None)
    self.assertEqual(p.active, True)

  def test_generator_implicit(self):

    """ Test a `Model`'s behavior when used as an iterator """

    # sample person
    p = TestPerson(firstname='John')

    # test implicit generator
    items = {}
    for name, value in p:
      items[name] = value

    self.assertEqual(len(items), 2)  # `active` should show up with default
    self.assertEqual(items['firstname'], 'John')
    self.assertEqual(items['active'], True)

  def test_generator_explicit(self):

    """ Test `Model` behavior when used as an iterator in `explicit` mode """

    # sample person
    p = TestPerson(firstname='John')

    # test explicit generator
    items = {}
    with p:
      for name, value in p:
        items[name] = value

    # should have _all_ properties
    self.assertEqual(len(items), len(p.__lookup__))
    self.assertEqual(items['firstname'], 'John')

    # defaults are returned as sentinels in `explicit` mode
    self.assertEqual(items['active'], datastructures.EMPTY)

    # unset properties are returned as sentinels in `explicit` mode
    self.assertEqual(items['lastname'], datastructures.EMPTY)

  def test_len(self):

    """ Test a `Model`'s behavior when used with `len()` """

    # sample person
    p = TestPerson()
    self.assertEqual(len(p), 0)

    # set 1st property
    p.firstname = 'John'
    self.assertEqual(len(p), 1)

    # set 2nd property
    p.lastname = 'Doe'
    self.assertEqual(len(p), 2)

  def test_nonzero(self):

    """ Test a `Model`'s falsyness with no properties """

    # sample peron
    p = TestPerson()
    self.assertTrue((not p))  # empty model should be falsy

    p.firstname = 'John'
    self.assertTrue(p)  # non-empty model is not falsy

  def test_get_invalid_property(self):

    """ Test getting an invalid `Model` property """

    # sample person
    p = TestPerson()

    # should be sealed off by `__slots__`
    with self.assertRaises(AttributeError):
      (p.blabble)

    # should be sealed off by metaclass-level `__slots__`
    with self.assertRaises(AttributeError):
      (TestPerson.blabble)

    # should be sealed off by good coding practices (lolz)
    with self.assertRaises(AttributeError):
      p._get_value('blabble')

  def test_get_value_all_properties(self):

    """ Test getting *all* properties via `_get_value` """

    # sample person
    p = TestPerson(firstname='John', lastname='Doe')
    properties = p._get_value(None)

    # should be a list of tuples
    self.assertEqual(len(properties), 6)
    self.assertIsInstance(properties, list)
    self.assertIsInstance(properties[0], tuple)

    # should retrieve even unset properties
    # (but they should be set to `None`, not the EMPTY sentinel of course)
    for k, v in properties:
      self.assertEqual(v, getattr(p, k))

  def test_model_getitem_setitem(self):

    """ Test a `Model`'s compliance with Python's Item API """

    # sample person
    p = TestPerson(firstname='John')

    # test __getitem__
    self.assertEqual(p.firstname, 'John')
    self.assertEqual(p['firstname'], 'John')

    # test __setitem__
    p['lastname'] = 'Gammon'
    self.assertEqual(p.lastname, 'Gammon')

    # try getting a nonexistent property, which should raise
    # `KeyError` instead of `AttributeError`
    with self.assertRaises(KeyError):
      (p['invalidproperty'])

    # make sure `AttributeError` still works properly
    with self.assertRaises(AttributeError):
      (p.invalidproperty)

  def test_model_setvalue(self):

    """ Test protected method `_set_value`, which is used by model internals """

    # sample person
    p = TestPerson(firstname='John')

    # try writing a new key
    x = p._set_value('key', model.VertexKey(TestPerson, "john"))
    self.assertEqual(x, p)

    # try writing to invalid property
    with self.assertRaises(AttributeError):
      p._set_value('invalidproperty', 'value')

    # quick test via descriptor API, which should also raise `AttributeError`
    with self.assertRaises(AttributeError):
      TestPerson.__dict__['firstname'].__set__(None, 'invalid')

  def test_model_setkey(self):

    """ Test protected method `_set_key`, which is used by model internals """

    # sample person
    p = TestPerson(firstname='John')

    # try writing an invalid key
    with self.assertRaises(TypeError):
      p._set_key(5.5)

    # try writing via kwargs
    k = model.VertexKey(TestPerson, "john")

    # try constructing via urlsafe
    p._set_key(urlsafe=k.urlsafe())

    # try constructing via raw
    p._set_key(raw=k.flatten(False)[1])

    # try already-constructed via kwargs
    p._set_key(constructed=k)

    # try providing both a value and formats, which should fail
    with self.assertRaises(TypeError):
      p._set_key(k, urlsafe=k.urlsafe())

    # try providing multiple formats, which should fail
    with self.assertRaises(TypeError):
      p._set_key(urlsafe=k.urlsafe(), raw=k.flatten(False)[1])

    # try passing nothing, which should fail
    with self.assertRaises(TypeError):
      p._set_key(None)

  def test_early_mutate(self):

    """ Test setting attributes and items on a model before it's ready """


    class EarlyMutateModel(model.Model):

      """ Tests mutation of properties before instantiation. """

      string = basestring

    # try writing to existing property
    with self.assertRaises(AttributeError):
      EarlyMutateModel.string = "testing123"

    # try writing a new property
    with self.assertRaises(AttributeError):
      EarlyMutateModel.newprop = "newvalue"

    # try writing to an internal property, which should work
    EarlyMutateModel.__impl__ = {}

  def test_validation_of_required_properties(self):

    """ Test validation of required properties """

    ## RequiredPropertyModel
    # Try validating properties marked as required.
    class RequiredPropertyModel(model.Model):

      """ Tests required properties. """

      nonrequired = basestring
      required = basestring, {'required': True}

    # sample model
    p = RequiredPropertyModel(nonrequired='sup')

    # try putting, should raise `ValueError`
    with self.assertRaises(ValueError):
      p.put()

    # should not raise errors
    p.required = 'sup'
    p.put()

  def test_validation_of_property_basetype(self):

    """ Test validation of property basetypes """

    ## BasetypedPropertyModel
    # Try validating properties by basetype.
    class BasetypedPropertyModel(model.Model):

      """ Tests property basetypes. """

      string = basestring
      number = int
      floating = float
      boolean = bool
      always_empty = basestring

    # sample model
    b = BasetypedPropertyModel()

    # test strings
    b.string = 5
    with self.assertRaises(ValueError):
      b.put()
    b.string = 'sample'

    # test integers
    b.number = '5'
    with self.assertRaises(ValueError):
      b.put()
    b.number = 5

    # test floats
    b.floating = 5
    with self.assertRaises(ValueError):
      b.put()
    b.floating = 5.5

    # test booleans
    b.boolean = 5.5
    with self.assertRaises(ValueError):
      b.put()
    b.boolean = True

    # should not except
    b.put()

  def test_validation_of_repeated_properties(self):

    """ Test validation of repeated properties """

    ## RepeatedPropertyModel
    # Try validating properties marked as repeated.
    class RepeatedPropertyModel(model.Model):

      """ Tests repeated properties. """

      nonrepeated = basestring
      repeated = int, {'repeated': True}

    # sample model
    r = RepeatedPropertyModel(nonrepeated=['blabble', '1', '2', '3'])

    # try repeated value in nonrepeated property
    with self.assertRaises(ValueError):
      r.put()

    # set to valid value
    r.nonrepeated = 'validvalue'

    # try nonrepeated value in repeated property
    r.repeated = 5
    with self.assertRaises(ValueError):
      r.put()

    # try invalid basetype in proper repeated field
    r.repeated = ['one', 'two', 'three']
    with self.assertRaises(ValueError):
      r.put()

    # set to valid value and put, should not except
    r.repeated = [1, 2, 3]
    r.put()

  def test_class_level_default_value(self):

    """ Test reading a property with a default set at the class level """

    ## ClassDefaultSample
    # Try grabbing the value of a defaulted property at the class level.
    class ClassDefaultSample(model.Model):

      """ Tests properties with default values at the class level. """

      sample_default = basestring, {'default': 'Hello, default!'}

    # try on the class level
    self.assertIsInstance(ClassDefaultSample.sample_default, model.Property)

  def test_class_level_propery_access(self):

    """ Test class-level property access on `Model` subclasses """

    assert isinstance(TestCar.make, model.Property)
    assert 'make' in TestCar.make.__repr__()
    assert 'Property' in TestCar.make.__repr__()

  def test_class_level_item_access(self):

    """ Test class-level item access on `Model` subclasses """

    assert isinstance(TestCar['make'], model.Property)
    assert 'make' in TestCar['make'].__repr__()
    assert 'Property' in TestCar['make'].__repr__()

  def test_graph_class_existence(self):

    """ Test proper export of ``Vertex`` and ``Edge`` models """

    assert hasattr(model, 'Vertex')
    assert hasattr(model, 'Edge')
    assert issubclass(model.Vertex, model.Model)
    assert issubclass(model.Edge, model.Model)

  def test_vertex_class_mro(self):

    """ Test proper MRO for ``Vertex`` models """

    assert hasattr(TestPerson, 'edges')
    assert hasattr(TestPerson(), 'edges')
    assert hasattr(TestPerson, 'neighbors')
    assert hasattr(TestPerson(), 'neighbors')
    assert issubclass(TestPerson, model.Vertex)
    assert isinstance(TestPerson(), model.Vertex)

  def test_edge_class_mro(self):

    """ Test proper MRO for ``Edge`` models """

    assert hasattr(Friendship, 'peers')
    assert issubclass(Friendship, model.Edge)



# -*- coding: utf-8 -*-

"""

  model adapter tests
  ~~~~~~~~~~~~~~~~~~~

  tests canteen's model adapters.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

if __debug__:

  # stdlib
  import os

  # canteen model API
  from canteen import model

  # canteen tests
  from canteen.test import FrameworkTest


  class AdapterExportTests(FrameworkTest):

    """ Tests objects exported by `model.adapter`. """

    def test_top_level_adapter_exports(self):

      """ Test that we can import concrete classes. """

      try:
        from canteen import model
        from canteen.model import adapter

      except ImportError:  # pragma: no cover
        return self.fail("Failed to import model adapter package.")

      else:
        self.assertIsInstance(adapter, type(os))  # `adapter` module
        self.assertIsInstance(adapter.abstract, type(os))  # `adapter.abstract`
        self.assertTrue(adapter.ModelAdapter)  # `ModelAdapter` parent class


  class ModelAdapterTests(FrameworkTest):

    """ Test `adapter.abstract.ModelAdapter`. """

    def test_adapter_registry(self):

      """ Test `adapter.abstract.ModelAdapter.registry`. """

      from canteen.model.adapter import abstract

      self.assertTrue(hasattr(abstract.ModelAdapter, 'registry'))
      self.assertIsInstance(abstract.ModelAdapter.registry, dict)

      ## SampleModel
      # Quick sample model to make sure class registration happens properly.
      class Sample(model.Model):

        """ Quick sample model. """

        pass

      # test that our class was registered
      self.assertTrue(('Sample' in abstract.ModelAdapter.registry))
      self.assertTrue(abstract.ModelAdapter.registry.get('Sample') == Sample)

    def test_default_adapter(self):

      """ Test that the default adapter is assigned properly. """

      from canteen.model import adapter
      from canteen.model.adapter import abstract

      ## TestDefault
      # Quick sample model to test default adapter injection.
      class TestDefault(model.Model):

        """ Quick sample model. """

        pass

      # test attribute + default injector
      self.assertTrue(hasattr(TestDefault, '__adapter__'))
      self.assertIsInstance(TestDefault.__adapter__, adapter.InMemoryAdapter)

    def test_explicit_adapter(self):

      """ Test that an adapter can be set explcitly on a model class. """

      from canteen.model import adapter
      from canteen.model.adapter import abstract

      ## TestExplicit
      # Quick sample model to test explicit adapter injection.
      class TestExplicit(model.Model):

        """ Quick sample model. """

        __adapter__ = adapter.RedisAdapter

      # test attribute + explicit injector
      self.assertTrue(hasattr(TestExplicit, '__adapter__'))
      self.assertIsInstance(TestExplicit.__adapter__, adapter.RedisAdapter)

    def test_mixin_repr(self):

      """ Test that a model `Mixin` generates a proper string representation """

      from canteen.model.adapter import core

      assert 'Mixin' in repr(core.AdaptedModel)
      assert 'AdaptedModel' in repr(core.AdaptedModel)


  __all__ = (
      'test_abstract',
      'test_inmemory',
      'test_protorpc',
      'test_redis')



# -*- coding: utf-8 -*-

"""

  abstract model adapters
  ~~~~~~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""


# stdlib
import abc
import time
import json
import base64
import logging
import datetime

# canteen utils
from canteen.util import config
from canteen.util import decorators


try:
  import zlib; _compressor = zlib
except ImportError:  # pragma: no cover
  pass  # pragma: no cover

try:
  import lz4; _compressor = lz4
except ImportError:  # pragma: no cover
  pass  # pragma: no cover

try:
  import snappy; _compressor = snappy
except ImportError:  # pragma: no cover
  pass  # pragma: no cover


## Globals
_adapters = {}
_adapters_by_model = {}
_compressor = None  # compressor for data marked for compression
_encoder = base64.b64encode  # encoder for key names and special strings
_core_mixin_classes = (
    'Mixin',
    'KeyMixin', 'ModelMixin',
    'CompoundKey', 'CompoundModel',
    'VertexMixin', 'EdgeMixin',
    'CompoundVertex', 'CompoundEdge')

## Computed Classes
CompoundKey = CompoundModel = CompoundVertex = CompoundEdge = None


class ModelAdapter(object):

  """ Abstract base class for classes that adapt canteen models to a particular
      storage backend. """

  registry = {}
  __metaclass__ = abc.ABCMeta

  @decorators.classproperty
  def config(cls):  # pragma: no cover

    """ Cached config shortcut.

        :returns: Configuration ``dict``, if any. Defaults
          to ``{'debug': True}``. """

    return config.Config().get(cls.__name__, {'debug': True})

  # noinspection PyMethodParameters
  @decorators.classproperty
  def logging(cls):

    """ Named logging pipe.

        :returns: Customized :py:mod:`canteen.util.debug.Logger`
          instance, with a ``name``/``path``/``condition``. """

    return logging  # @TODO(sgammon): proper logging

  @decorators.classproperty
  def serializer(cls):

    """ Load and return the appropriate serialization codec. This property is
        mainly meant to be overridden by child classes.

        :returns: Current serializer, defaults to :py:mod:`json`. """

    return json  # default to JSON

  @decorators.classproperty
  def encoder(cls):  # pragma: no cover

    """ Encode a stringified blob for storage. This property is mainly meant to
        be overridden by child classes.

        :returns: Current :py:data:`_encoder`, defaults to :py:mod:`base64`. """

    return _encoder  # use local encoder

  @decorators.classproperty
  def compressor(cls):  # pragma: no cover

    """ Load and return the appropriate compression codec. This property is
        mainly meant to be overridden by child classes.

        :returns: Current :py:mod:`_compressor`, defaults to :py:mod:`zlib`. """

    return _compressor

  ## == Internal Methods == ##
  def _get(self, key, **kwargs):

    """ Low-level method for retrieving an entity by Key. Fetches and
        deserializes the given entity, if it exists, or returns ``None``.

        :param key: Instance of :py:class:`model.Key` to retrieve from storage.

        :raises RuntimeError: If the target :py:class:`adapter.ModelAdapter`
          does not implement ``get()``, which is an ABC-enforced child class
          method. :py:exc:`RuntimeError` and descendents are also re-raised from
          the concrete adapter.

        :returns: Inflated :py:class:`model.Model` instance, corresponding to
            ``key``, or ``None`` if no such entity could be found. """

    if self.config.get('debug', False):  # pragma: no cover
      self.logging.info("Retrieving entity with Key: \"%s\"." % key)

    # immediately fail with no overriden `get`
    if not hasattr(self.__class__, 'get') and (
      self.__class__ != ModelAdapter):  # pragma: no cover
      ctx = self.__class__.__name__
      raise RuntimeError("ModelAdapter `%s` does not implement `get`,"
                         " and thus cannot be used for reads." % ctx)
    else:
      # grab getter method
      getter = getattr(self.__class__, 'get')

    # flatten key into stringified repr
    joined, flattened = key.flatten(True)
    parent, kind, id = flattened

    # optionally allow adapter to encode key
    encoded = self.encode_key(joined, flattened)

    if not encoded:  # pragma: no cover
      # otherwise, use regular base64 via `AbstractKey`
      encoded = key.urlsafe(joined)

    # pass off to delegated `get`
    try:
      entity = getter((encoded, flattened), **kwargs)
    except NotImplementedError:  # pragma: no cover
      ctx = self.__class__.__name__
      raise RuntimeError("ModelAdapter `%s` does not implement `get`,"
                         " and thus cannot be used for reads." % ctx)
    except RuntimeError:  # pragma: no cover
      raise
    else:
      if entity is None: return  # not found

      if isinstance(entity, dict):  # inflate dict
        entity['key'] = key
        entity = self.registry[kind](_persisted=True, **entity)

      # inflate key + model and return
      key.__persisted__ = True
      return entity

  def _put(self, entity, **kwargs):

    """ Low-level method for persisting an Entity. Collapses and serializes the
        target ``entity`` into native types and delegates to the active
        :py:class:`model.adapter.ModelAdapter` for storage.

        :param entity: Object descendent of :py:class:`model.Model`, suitable
          for storage via the currently-active adapter.

        :raises ValueError: In the case of an unknown or unregistered *kind*.
        :returns: New (or updated) key value for the target ``entity``. """

    # resolve model class
    _model = self.registry.get(entity.kind())
    if not _model:  # pragma: no cover
      raise ValueError('Could not resolve model class "%s".' % entity.kind())

    with entity:  # enter explicit mode

      # validate entity, will raise validation exceptions
      for name, value in entity.to_dict(_all=True).items():
        _model[name].valid(entity)

      # resolve key if we have a zero-y key or key class
      if not entity.key or entity.key is None:
        # build an ID-based key
        ids = self.allocate_ids(_model.__keyclass__, entity.kind())
        entity._set_key(_model.__keyclass__(entity.kind(), ids))

      # flatten key/entity
      joined, flat = entity.key.flatten(True)

    # delegate
    return self.put((
      self.encode_key(joined, flat) or entity.key.urlsafe(joined), flat),
        entity._set_persisted(True), _model, **kwargs)

  def _delete(self, key, **kwargs):

    """ Low-level method for deleting an entity by Key.

        :param key: Target :py:class:`model.Key` to delete.
        :returns: Result of the delete operation. """

    if self.config.get('debug', False):  # pragma: no cover
      self.logging.info("Deleting Key: \"%s\"." % key)

    joined, flat = key.flatten(True)
    return self.delete((
      self.encode_key(joined, flat) or key.urlsafe(joined), flat), **kwargs)

  @classmethod
  def _register(cls, model):

    """ Low-level method for registering a Model class with this adapter's this
        adapter's registry.

        :param model: :py:class:`model.Model` class to register.
        :returns: The ``model`` it was handed (for chainability). """

    cls.registry[model.kind()] = model
    return model

  ## == Class Methods == ##
  @classmethod
  def acquire(cls, name, bases, properties):

    """ Acquire a new/existing copy of this adapter. Available for override by
        child classes to customize the driver acquisition process. Passed an
        identical signature to ``type``, *before* the :py:class:`model.Model`
        class has been fully-built.

        :param name: String name of the new :py:class:`model.Model` class-to-be.
        :param bases: Tuple of base classes for the target :py:class:`Model`.
        :param properties: Property ``dict`` from class definition.

        :returns: The "acquired" adapter object. """

    global _adapters
    global _adapters_by_model

    # if we don't have one yet, spawn a singleton
    if cls.__name__ not in _adapters:
      _adapters[cls.__name__] = cls()
    _adapters_by_model[name] = _adapters[cls.__name__]
    return _adapters[cls.__name__]

  ## == Abstract Methods == ##
  @abc.abstractmethod
  def get(cls, key, **kwargs):  # pragma: no cover

    """ Retrieve an entity by :py:class:`model.Key`. Must accept a tuple in the
        format ``(<joined Key repr>, <flattened key>)``. Abstract method that
        **must** be overridden by concrete implementors of
        :py:class:`ModelAdapter`.

        :param key: Target :py:class:`model.Key` to retrieve.
        :raises: :py:exc:`NotImplementedError`, as this method is abstract. """

    raise NotImplementedError('`ModelAdapter.get`'
                              ' is abstract and may not be'
                              ' called directly.')  # pragma: no cover

  @abc.abstractmethod
  def put(cls, key, entity, model, **kwargs):  # pragma: no cover

    """ Persist an entity in storage. Must accept a :py:class:`model.Key`, which
        may not have an ID, in which case one is allocated. The entity and
        :py:class:`model.Model` class are also passed in.

        This method is abstract and **must** be overridden by concrete
        implementors of :py:class:`ModelAdapter`.

        :param key: Potentially-empty :py:class:`model.Key` for the new entity.
        :param entity: Object :py:class:`model.Model` to persist in storage.
        :param model: :py:class:`model.Model` class for target ``entity``.
        :raises: :py:exc:`NotImplementedError`, as this method is abstract. """

    raise NotImplementedError('`ModelAdapter.put`'
                              ' is abstract and may not be'
                              ' called directly.')  # pragma: no cover

  @abc.abstractmethod
  def delete(cls, key, **kwargs):  # pragma: no cover

    """ Delete an entity by :py:class:`model.Key`. Must accept a target
        ``key``, whose associated entity will be deleted.

        This method is abstract and **must** be overridden by concrete
        implementors of :py:class:`ModelAdapter`.

        :param key: Target :py:class:`model.Key` to delete.
        :raises: :py:exc:`NotImplementedError`, as this method is abstract. """

    raise NotImplementedError('`ModelAdapter.delete`'
                              ' is abstract and may not be'
                              ' called directly.')  # pragma: no cover

  @abc.abstractmethod
  def allocate_ids(cls, key_cls, kind, count=1, **kwargs):  # pragma: no cover

    """ Allocate new :py:class:`model.Key` IDs for ``kind`` up to ``count``.
        This method is abstract and **must** be overridden by concrete
        implementors of :py:class:`ModelAdapter`.

        :param key_class: :py:class:`model.Key` class for provisioned IDs.
        :param kind: String ``kind`` name from :py:class:`model.Model` class.
        :param count: Count of IDs to provision, defaults to ``1``.
        :raises: :py:exc:`NotImplementedError`, as this method is abstract. """

    raise NotImplementedError('`ModelAdapter.allocate_ids`'
                              ' is abstract and may not be'
                              ' called directly.')  # pragma: no cover

  @classmethod
  def encode_key(cls, key, joined, flattened):

    """ Encode a :py:class:`model.Key` for storage. This method is abstract and
        *should* be overridden by concrete implementors of
        :py:class:`ModelAdapter`.

        In the case that a :py:class:`ModelAdapter` wishes to defer to the
        default encoder (:py:mod:`base64`), it can return ``False``.

        :param key: Target :py:class:`model.Key` to encode.
        :param joined: Joined/stringified key.
        :param flattened: Flattened ``tuple`` (raw) key.
        :returns: The encoded :py:class:`model.Key`, or ``False`` to
              yield to the default encoder. """

    # by default, yield to key b64 builtin encoding
    return False  # pragma: no cover


# noinspection PyAbstractClass
class IndexedModelAdapter(ModelAdapter):

  """ Abstract base class for model adapters that support additional indexing
      APIs. """

  # magic prefixes
  _key_prefix = '__key__'
  _kind_prefix = '__kind__'
  _group_prefix = '__group__'
  _index_prefix = '__index__'
  _reverse_prefix = '__reverse__'


  class Indexer(object):

    """ Holds methods for indexing and handling index data types. """

    _magic = {
      'key': 0x1,  # magic ID for `model.Key` references
      'date': 0x2,  # magic ID for `datetime.date` instances
      'time': 0x3,  # magic ID for `datetime.time` instances
      'datetime': 0x4}  # magic ID for `datetime.datetime` instances

    @classmethod
    def convert_key(cls, key):

      """ Convert a :py:class:`model.Key` to an indexable value.

          :param key: Target :py:class:`model.Key` to convert.

          :returns: Tupled ``(<magic key code>, <flattened key>)``, suitable for
            adding to the index. """

      # flatten and return key structure with magic
      joined, flattened = key.flatten(True)
      return cls._magic['key'], map(lambda x: x is not None, flattened)

    @classmethod
    def convert_date(cls, _date):

      """ Convert a Python ``date`` to an indexable value.

          :param date: Python ``date`` to convert.

          :returns: Tupled ``(<magic date code>, <flattened date>)`` to add to
            the index. """

      # convert to ISO format, return date with magic
      return cls._magic['date'], int(time.mktime(_date.timetuple()))

    @classmethod
    def convert_time(cls, _time):

      """ Convert a Python ``time`` to an indexable value.

          :param _time: Python ``time`` to convert.

          :returns: Tupled ``(<magic time code>, <flattened time>)``, suitable
            for addition to the index. """

      # convert to ISO format, return time with magic
      return cls._magic['time'], _time.isoformat()

    @classmethod
    def convert_datetime(cls, _datetime):

      """ Convert a Python ``datetime`` to an indexable value.

          :param _datetime: Python ``datetime`` to convert.

          :returns: Tupled ``(<magic time code>, <flattened datetime>)``,
            suitable for addition to the index. """

      # convert to integer, return datetime with magic
      return cls._magic['datetime'], int(time.mktime(_datetime.timetuple()))

  @decorators.classproperty
  def _index_basetypes(cls):

    """ Map basetypes to indexer routines.

        :returns: Default basetype ``dict``. """

    from canteen import model

    return {

      # -- basetypes -- #
      int: cls.serializer.dumps,
      bool: cls.serializer.dumps,
      long: cls.serializer.dumps,
      float: cls.serializer.dumps,
      basestring: cls.serializer.dumps,

      # -- model/key types -- #
      model.Key: cls.Indexer.convert_key,

      # -- date/time types -- #
      datetime.date: cls.Indexer.convert_date,
      datetime.time: cls.Indexer.convert_time,
      datetime.datetime: cls.Indexer.convert_datetime

    }

  def _put(self, entity, **kwargs):  # pragma: no cover

    """ Hook to trigger index writes for a given entity. Defers up the chain to
        :py:class:`ModelAdapter` after generating (and potentially writing) a
        set of indexes from the target ``entity``.

        :param entity: Entity :py:class:`model.Model` to persist.

        :returns: Resulting :py:class:`model.Key` from write operation. """

    # small optimization - with a deterministic key, we can parrellelize
    # index writes (assuming async is supported in the underlying driver)

    _indexed_properties = self._pluck_indexed(entity)

    # delegate write up the chain
    written_key = super(IndexedModelAdapter, self)._put(entity, **kwargs)

    # proxy to `generate_indexes` and write indexes
    if not _indexed_properties:  # pragma: no cover
      origin, meta = self.generate_indexes(entity.key)
      property_map = {}
    else:
      origin, meta, property_map = (
        self.generate_indexes(entity.key, _indexed_properties))

    self.write_indexes((origin, meta, property_map), **kwargs)
    return written_key  # delegate up the chain for entity write

  def _delete(self, key, **kwargs):

    """ Hook to trigger index cleanup for a given key. Defers up the chain to
        :py:class:`ModelAdapter` after generating a set of indexes to clean for
        the target ``key``.

        :param key: Target :py:class:`model.Key` to delete.

        :returns: Result of delete operation. """

    # generate meta indexes only, then clean
    self.clean_indexes(self.generate_indexes(key))

    # delegate delete up the chain
    return super(IndexedModelAdapter, self)._delete(key)

  @staticmethod
  def _pluck_indexed(entity, context=None, _map=None):

    """ Zip and pluck only properties that should be indexed. Simply returns a
        set of property descriptors, mapped to ehtir names in a ``dict``, if
        they are marked as needing to be indexed.

        :param entity: Target entity to produce indexes for.

        :returns: Map ``dict`` of properties to index. """

    from canteen import model

    _map = _map or {}
    _edict = entity if isinstance(entity, dict) else entity.to_dict(
      convert_datetime=False)

    # grab only properties enabled for indexing
    is_indexed = lambda x: entity.__class__.__dict__[x[0]]._indexed
    for k, v in filter(is_indexed, _edict.items()):
      prop = entity.__class__.__dict__[k]

      if isinstance(prop._basetype, type) and (
              issubclass(prop._basetype, model.Model) and (
              prop._options.get('embedded'))):  # pragma: no cover
        _map = IndexedModelAdapter._pluck_indexed(getattr(entity, k), k, _map)

      else:
        # attach property name, property class, value
        _map['.'.join((context, k)) if context else k] = (prop, v)

    return _map

  def _execute_query(self, query):

    """ Execute a ``query.Query`` object, returning results that match the
        search terms specified in ``query`` and the attached
        ``query.QueryOptions`` object.

        :param query: ``query.Query`` to execute via the local adapter.

        :returns: Query results, if any. """

    return self.execute_query(*(
      query.kind, (query.filters, query.sorts), query.options))

  @classmethod
  def generate_indexes(cls, key, properties=None):

    """ Generate a set of indexes that should be written to with associated
        values.

        :param key: Target :py:class:`model.Key` to index.
        :param properties: Entity :py:class:`model.Model` property values to
          index.

        :returns: Tupled set of ``(encoded, meta, property)``, where ``meta``
          and ``property`` are indexes to be written in each category. """

    if key is None and not properties:  # pragma: no cover
      raise TypeError('Must pass at least `key` or `properties'
                      ' to `generate_indexes`.')

    _property_indexes, _meta_indexes = [], []

    if key is not None:

      # provision vars, generate meta indexes
      # meta indexes look like:
      #  (`__key__`,), target
      #  (`__kind__`, kind), target

      encoded_key = cls.encode_key(*key.flatten(True)) or key.urlsafe()
      _meta_indexes.append((cls._key_prefix,))
      _meta_indexes.append((cls._kind_prefix, key.kind))  # map kind

      # consider ancestry
      if not key.parent:

        # generate group indexes in the case of a nonvoid parent
        # group indexes look like:
        #  (`<group-key>`,), target

        _meta_indexes.append((cls._group_prefix,))

      else:

        # generate group prefix for root
        # root indexes look like:
        #  `__group__`, trimmed-target-root-key

        # append keyparent-based group prefix
        root_key = (i for i in key.ancestry).next()

        # encode root key
        encoded_root_key = (
          cls.encode_key(*root_key.flatten(True)) or root_key.urlsafe())
        _meta_indexes.append((cls._group_prefix, encoded_root_key))

    # add property index entries
    if properties is not None:

      # we're applying writes
      for k, v in properties.items():

        # extract property class and value
        prop, value = v

        # consider repeated properties
        if not prop.repeated or not isinstance(value, (
          tuple, list, set, frozenset)):
          value = [value]

        # generate property index entries for values
        # property value indexes look like:
        #  `__index__::kind::property::encoded_value`, target

        # and the internal representation looks like:
        #  `(value_encoder_callable, (index_prefix, kind, propname, value))`

        # iterate through property values
        for x in value:
          context = (cls._index_prefix, key.kind, k, x)
          _property_indexes.append((
            cls._index_basetypes.get(prop.basetype, basestring), context))

        continue

    else:
      # we're cleaning indexes
      return encoded_key, _meta_indexes
    return encoded_key, _meta_indexes, _property_indexes

  @abc.abstractmethod
  def write_indexes(cls, writes, **kwargs):

    """ Write a batch of index updates generated earlier via
        :py:meth:`generate_indexes`. This method is abstract and **must** be
        overridden by concrete implementors of :py:class:`IndexedModelAdapter`.

        :param writes: Batch of index writes to commit, generated via
          :py:meth:`generate_indexes`.

        :raises: :py:exc:`NotImplementedError`, as this method is abstract. """

    raise NotImplementedError('`IndexedModelAdapter.write_indexes`'
                              ' is abstract and may not be'
                              ' called directly.')  # pragma: no cover

  @abc.abstractmethod
  def clean_indexes(cls, key, **kwargs):

    """ Clean indexes and index entries matching a particular
        :py:class:`model.Key`. This method is abstract and **must** be
        overridden by concrete implementors of :py:class:`IndexedModelAdapter`.

        :param key: Target :py:class:`model.Key` to clean indexes for.
        :raises: :py:exc:`NotImplementedError`, as this method is abstract. """

    raise NotImplementedError('`IndexedModelAdapter.clean_indexes`'
                              ' is abstract and may not be'
                              ' called directly.')  # pragma: no cover

  @abc.abstractmethod
  def execute_query(cls, kind, spec, options, **kwargs):

    """ Execute a query, specified by ``spec``, across one (or multiple) indexed
        properties.

        :param spec: Object specification (:py:class:`model.Query`) specifying
          the query to satisfy.

        :raises: :py:exc:`NotImplementedError`, as this method is abstract. """

    raise NotImplementedError('`IndexedModelAdapter.execute_query`'
                              ' is abstract and may not be'
                              ' called directly.')  # pragma: no cover


class GraphModelAdapter(IndexedModelAdapter):

  """ Abstract base class for model adapters that support Graph-style paradigms
      for data storage. """

  # graph/vertex/edge prefixes
  _edge_prefix = '__edge__'
  _graph_prefix = '__graph__'
  _vertex_prefix = '__vertex__'

  # universal tokens
  _neighbors_token = 'neighbors'

  # directed tokens
  _in_token = 'in'
  _out_token = 'out'
  _directed_token = 'directed'

  # undirected tokens
  _peers_token = 'peers'
  _undirected_token = 'undirected'


  class Indexer(IndexedModelAdapter.Indexer):

    """ Adds Graph-specific ``Indexer`` routines and constants. """

    _magic = {
      'key': 0x1,  # magic ID for `model.Key` references
      'date': 0x2,  # magic ID for `datetime.date` instances
      'time': 0x3,  # magic ID for `datetime.time` instances
      'datetime': 0x4,  # magic ID for `datetime.datetime` instances
      'vertex': 0x5,  # magic ID for `Vertex` `model.Key` references
      'edge': 0x6}  # magic ID for `Edge` `model.Key` references

    @classmethod
    def convert_key(cls, key):

      """ Convert a :py:class:`model.Key` to an indexable value, considering
          ``Vertex`` and ``Edge`` keys as well.

          :param key: Target :py:class:`model.Key` to convert.

          :returns: Tupled ``(<magic key code>, <flattened key>)``, suitable for
            adding to the index. """

      from canteen import model

      if isinstance(key, basestring):
        # expand from urlsafe
        # @TODO(sgammon): this is disgusting what was i thinking
        # @TODO(sgammon): come up with a proper native repr for a key
        key = model.Key(urlsafe=key)  # pragma: no cover

      joined, flattened = key.flatten(True)
      sanitized = map(lambda x: x is not None, flattened)

      _GRAPH_KEYS = (model.Vertex.__keyclass__, model.Edge.__keyclass__)

      _key = 'key'
      if hasattr(key, '__vertex__') and key.__vertex__:
        _key = 'vertex'  # pragma: no cover
      elif hasattr(key, '__edge__') and key.__edge__:
        _key = 'edge'  # pragma: no cover
      return (cls._magic[_key], key.urlsafe())

  @decorators.classproperty
  def _index_basetypes(self):

    """ Map basetypes to indexer routines, with support for graph-specialized
        key types (``VertexKey`` and ``EdgeKey``).

        :returns: Default basetype ``dict``. """

    from canteen import model
    types = super(GraphModelAdapter, self)._index_basetypes

    types.update({
      # -- graph key types -- #
      model.Key: self.Indexer.convert_key,
      model.EdgeKey: self.Indexer.convert_key,
      model.VertexKey: self.Indexer.convert_key})
    return types

  @staticmethod
  def _pluck_indexed(entity):

    """ Override the indexed property scanner to ignore ``Edge``-related auto-
        injected properties.

        :param entity: Target entity to produce indexes for.

        :returns: Map ``dict`` of properties to index. """

    _map = IndexedModelAdapter._pluck_indexed(entity)

    # don't index graph properties - they are resovled via graph indexes
    if hasattr(entity.__class__, '__edge__') and entity.__class__.__edge__:
      if 'peers' in _map: del _map['peers']
      if 'target' in _map: del _map['target']
      if 'source' in _map: del _map['source']
    return _map

  def _put(self, entity, **kwargs):

    """ Override to enable ``graph``-specific indexes (for stored ``Vertex`` and
        ``Edge`` objects/keys).

        :param entity: Entity :py:class:`model.Model` to persist.

        :returns: Resulting :py:class:`model.Key` from write operation. """

    # small optimization - with a deterministic key, we can parrellelize
    # index writes (assuming async is supported in the underlying driver)

    _indexed_properties = self._pluck_indexed(entity)

    # delegate write up the chain
    written_key = super(IndexedModelAdapter, self)._put(entity, **kwargs)

    # proxy to `generate_indexes` and write indexes
    origin, meta, properties, graph = (
      self.generate_indexes(entity.key, entity, _indexed_properties))

    self.write_indexes((origin, meta, properties), graph, **kwargs)
    return written_key  # delegate up the chain for entity write

  @classmethod
  def generate_indexes(cls, key, entity=None, properties=None):

    """ Generate a set of indexes that should be written to with associated
        values, considering that some ``key`` values may be ``VertexKey`` or
        ``EdgeKey`` instances.

        :param key: Target :py:class:`model.Key`, :py:class:`VertexKey` or
          :py:class:`EdgeKey` to index.

        :param entity: :py:class:`Model` entity to be stored. Defaults to
          ``None``, in which case we're cleaning indexes and don't have access
          to the original entity - just the key.

        :param properties: Entity :py:class:`model.Model` property values to
          index.

        :raises TypeError: If neither a ``key`` or ``properties`` are passed,
          since we can't generate anything without at least one or the other.

        :returns: Tupled set of ``(encoded, meta, property, graph)``, where
          ``meta`` and ``property`` are indexes to be written in each category
           and ``graph`` is a bundle of special indexes for ``Vertex`` and
           ``Edge`` keys. """

    from .. import Key, Model

    if key is None and not properties:  # pragma: no cover
      raise TypeError('Must pass at least `key` or `properties'
                      ' to `generate_indexes`.')

    if not (entity is None or isinstance(entity, Model)):  # pragma: no cover
      raise TypeError('Must pass either `None` or a `Model`'
                      ' for the `entity` parameter to `generate_indexes`.'
                      ' Instead, got: "%s".' % entity)

    from .. import VertexKey, EdgeKey
    _super = super(GraphModelAdapter, cls).generate_indexes

    # initialize graph indexes
    graph = []

    if key and properties is None:
      # we're probably cleaning indexes
      encoded, meta = _super(key)

    else:
      # defer upwards for regular indexes
      encoded, meta, properties = _super(key, properties)

    if key:
      # vertex keys
      if isinstance(key, VertexKey):
        graph.append((cls._vertex_prefix,))

      # edge keys
      elif isinstance(key, EdgeKey):

        # main edge index
        graph.append((cls._edge_prefix,))

        if entity:
          # extract spec @TODO(sgammon): make specs not suck
          spec = entity.__spec__

          # directed/undirected index
          graph.append((cls._edge_prefix, cls._directed_token if (
            spec.directed) else cls._undirected_token))

          # directed indexes
          if spec.directed:

            for target in entity.target:

              if isinstance(target, Key) and not (
                  isinstance(target, VertexKey)):  # pragma: no cover
                # @TODO(sgammon): unambiguous graph keys
                target = VertexKey.from_urlsafe(target.urlsafe())

              # __graph__::<source>::out => edge
              graph.append((entity.source, cls._out_token, entity.key))

              # __graph__::<target>::in => edge
              graph.append((target, cls._in_token, entity.key))

              # __graph__::<source>::neighbors => target
              graph.append((entity.source, cls._neighbors_token, target))

              # __graph__::<target>::neighbors => source
              graph.append((target, cls._neighbors_token, entity.source))

          # undirected indexes
          else:

            _indexed_pairs = set()
            for o, source in enumerate(entity.peers):

              if isinstance(source, Key) and not (
                  isinstance(source, VertexKey)):  # pragma: no cover
                # @TODO(sgammon): unambiguous graph keys
                source = VertexKey.from_urlsafe(source.urlsafe())

              for i, target in enumerate(entity.peers):

                if isinstance(target, Key) and not (
                    isinstance(target, VertexKey)):  # pragma: no cover
                  # @TODO(sgammon): unambiguous graph keys
                  target = VertexKey.from_urlsafe(target.urlsafe())

                # skip if it's the same object in the pair
                if o == i: continue  # pragma: no cover

                # skip if we've already indexed the two, since we're undirected
                # and one iteration past either will work for both
                if (source, target) in _indexed_pairs or (
                    (target, source) in _indexed_pairs):  # pragma: no cover
                  continue

                # __graph__::<source>::peers => edge
                graph.append((source, cls._peers_token, entity.key))

                # __graph__::<target>::peers => edge
                graph.append((target, cls._peers_token, entity.key))

                # __graph__::<source>::neighbors => target
                graph.append((source, cls._neighbors_token, target))

                # __graph__::<target>::neighbors => source
                graph.append((target, cls._neighbors_token, source))

    if key and properties is None:
      return encoded, meta, tuple(graph)
    return encoded, meta, properties, tuple(graph)

  @abc.abstractmethod
  def write_indexes(cls, writes, graph, **kwargs):

    """ Write a batch of index updates generated earlier via
        :py:meth:`generate_indexes`. This method is abstract and **must** be
        overridden by concrete implementors of :py:class:`IndexedModelAdapter`.

        :param writes: Batch of index writes to commit, generated via
          :py:meth:`generate_indexes`.

        :param graph: Batch of graph-related storage updates for indexes related
          to ``Edge`` and ``Vertex`` objects.

        :raises: :py:exc:`NotImplementedError`, as this method is abstract. """

    raise NotImplementedError('`GraphModelAdapter.write_indexes`'
                              ' is abstract and may not be'
                              ' called directly.')  # pragma: no cover

  @abc.abstractmethod
  def clean_indexes(cls, key, graph, **kwargs):

    """ Clean indexes and index entries matching a particular
        :py:class:`model.Key`. This method is abstract and **must** be
        overridden by concrete implementors of :py:class:`IndexedModelAdapter`.

        :param key: Target :py:class:`model.Key` to clean indexes for.
        :raises: :py:exc:`NotImplementedError`, as this method is abstract. """

    raise NotImplementedError('`IndexedModelAdapter.clean_indexes`'
                              ' is abstract and may not be'
                              ' called directly.')  # pragma: no cover


class DirectedGraphAdapter(GraphModelAdapter):

  """ Abstract base class for model adpaters that support directed-graph-type
      models. """


# noinspection PyAttributeOutsideInit
class Mixin(object):

  """ Abstract parent for detecting and registering `Mixin` classes. """

  __slots__ = tuple()

  class __metaclass__(type):

    """ Local `Mixin` metaclass for registering encountered `Mixin`(s). """

    ## == Mixin Registry == ##
    _compound = {}
    _mixin_lookup = set()
    _key_mixin_registry = {}
    _model_mixin_registry = {}
    _vertex_mixin_registry = {}
    _edge_mixin_registry = {}

    def __new__(cls, name, bases, properties):

      """ Factory a new registered :py:class:`Mixin`. Registers the target
          ``Mixin`` in :py:attr:`Mixin.__metaclass__._mixin_lookup`, and
          extends compound class at :py:attr:`Mixin.__metaclass__._compound`.

          :param name: Name of ``Mixin`` class to construct.
          :param bases: Class bases of ``Mixin`` class to construct.
          :param properties: Mapping ``dict`` of class properties.
          :raises RuntimeError: For invalid inheritance between mixin bases.

          :returns: Constructed ``Mixin`` class. """

      # apply local metaclass to factoried concrete children
      klass = super(cls, cls).__new__(cls, name, bases, properties)

      # register mixin if it's not a concrete parent and is unregistered
      if name not in frozenset(_core_mixin_classes) and (
        name not in cls._mixin_lookup):

        # add to each registry that the mixin supports
        for base in bases:

          ## add mixin to parent registry
          base.__registry__[name] = klass

        # add to global mixin lookup to prevent double loading
        cls._mixin_lookup.add(name)

        # see if we already have a compound class (mixins loaded after models)
        if Mixin._compound.get(cls):

          ## extend class dict if we already have one
          Mixin._compound.__dict__.update(*(
            dict(cls.__dict__.items())))  # pragma: no cover

      return klass

    def __repr__(cls):

      """ Generate a string representation of a `Mixin` subclass.

          :returns: String *repr* for ``Mixin`` class. """

      return "Mixin(%s.%s)" % (cls.__module__, cls.__name__)

  internals = __metaclass__

  @decorators.classproperty
  def methods(cls):

    """ Recursively return all available ``Mixin`` methods.

        :yields: Each method in each ``Mixin``. """

    for component in cls.components:
      for method, func in component.__dict__.items():
        yield method, func

  @decorators.classproperty
  def compound(cls):

    """ Generate a compound ``Mixin`` class. Builds a new class, composed of all
        available methods on attached mixins.

        :returns: Factoried compound ``Mixin`` class. """

    global CompoundKey, CompoundModel, CompoundVertex, CompoundEdge

    if isinstance(cls.__compound__, basestring):

      # if we've never generated a `CompoundModel`, regenerate...
      cls.__compound__ = cls.internals._compound[cls] = type(*(
        cls.__compound__,
        (cls, object),
        dict([
          ('__origin__', cls),
          ('__slots__', tuple()),
        ] + [(k, v) for k, v in cls.methods])))

      if cls.__compound__.__name__ == 'CompoundKey':
        CompoundKey = cls.__compound__
      elif cls.__compound__.__name__ == 'CompoundModel':
        CompoundModel = cls.__compound__
      elif cls.__compound__.__name__ == 'CompoundVertex':
        CompoundVertex = cls.__compound__
      elif cls.__compound__.__name__ == 'CompoundEdge':
        CompoundEdge = cls.__compound__

    return cls.__compound__

  @decorators.classproperty
  def components(cls):

    """ Return registered ``Mixin`` classes for the current ``cls``.

        :yields: Each mixin in the registry. """

    for mixin in cls.__registry__.itervalues(): yield mixin


class KeyMixin(Mixin):

  """ Allows injection of attributes into `Key`. """

  __slots__ = tuple()
  __compound__ = 'CompoundKey'
  __registry__ = Mixin._key_mixin_registry


class ModelMixin(Mixin):

  """ Allows injection of attributes into `Model`. """

  __slots__ = tuple()
  __compound__ = 'CompoundModel'
  __registry__ = Mixin._model_mixin_registry


class VertexMixin(Mixin):

  """ Allows injection of attributes into `Vertex`. """

  __slots__ = tuple()
  __compound__ = 'CompoundVertex'
  __registry__ = Mixin._vertex_mixin_registry


class EdgeMixin(Mixin):

  """ Allows injection of attributes into `Edge`. """

  __slots__ = tuple()
  __compound__ = 'CompoundEdge'
  __registry__ = Mixin._edge_mixin_registry



# -*- coding: utf-8 -*-

"""

  realtime logic
  ~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import time
import hashlib

# core and utils
from ..base import logic
from ..util import struct
from ..util import decorators


# Globals
_ORIGIN_ENV_ITEM = 'HTTP_ORIGIN'
_SOCKET_KEY_ENV_ITEM = 'HTTP_SEC_WEBSOCKET_KEY'
_FORWARDED_FOR_ENV_ITEM = 'HTTP_X_FORWARDED_FOR'
TERMINATE = struct.Sentinel('TERMINATE')


class TerminateSocket(Exception):

  """ Exception that can be raised from the inner receive/send coroutine to
      gracefully (or non-gracefully) close an active realtime communication
      session. """

  __graceful__ = False  # was this shutdown requested or forced?

  def __init__(self, graceful=False):

    """ Initialize this ``TerminateSocket`` exception.

        :param graceful: ``bool`` flag indicating whether to gracefully close
          the connection or force a close. ``True`` tries to close the socket
          gracefully, ``False`` does not. """

    self.__graceful__ = graceful

  # ~~ accessors ~~ #
  graceful = property(lambda self: self.__graceful__)


class RealtimeSocket(object):

  """ Specifies a structure for keeping track of socket details during a
      live realtime communication session. Simple things like the local/remote
      address/port pairs are used to make a socket ID that can be passed around
      and used as a reference. """

  __slots__ = (
    '__id__',  # ID of the socket
    '__state__',  # current state
    '__local__',  # local port pair
    '__remote__',  # remote port pair
    '__runtime__',  # link to current runtime
    '__established__')  # establish timestamp

  class State(struct.BidirectionalEnum):

    """ Enumerates available ``RealtimeSocket`` states, or the phases of a
        realtime communication session. """

    INIT = 0x0  # socket has just been initialized
    OPEN = 0x1  # socket is open but has not yet sent or received data
    ACTIVE = 0x2  # socket has sent or received data and is currently active
    CLOSED = 0x3  # socket was closed gracefully by the client or server
    ERROR = 0x4  # socket was closed in an error state

  def __init__(self, runtime, local, remote):

    """ Initialize this ``RemoteSocket`` object with details about the current
        ``runtime`` and the connection's ``local`` and ``remote`` peers.

        :param runtime: Currently-active Canteen runtime.

        :param local: Local (server) address & port pair that is handling the
          connection.

        :param remote: Remote (client) address & port pair that is on the other
          side of the connection. """

    self.__local__, self.__remote__ = local, remote
    self.__id__, self.__runtime__, self.__established__ = (
      hashlib.sha1('::'.join((local, remote))).hexdigest(),
      runtime,
      int(time.time()))

  def __hash__(self):

    """ Return the local socket's ID.

        :returns: This ``RealtimeSocket``'s ID for hashability. """

    return self.__id__

  def __repr__(self):

    """ Return a humanized string representation of this ``RealtimeSocket``.

        :returns: This ``RealtimeSocket``'s humanized ``repr``. """

    return "%s(id=%s, state=%s, local=%s, remote=%s, established=%s)" % (
      self.__class__.__name__,
      self.__id__,
      self.State.reverse_resolve(self.__state__),
      self.__local__,
      self.__remote__,
      self.__established__)

  def set_state(self, state):

    """ Set the current state of this ``RealtimeSocket``. Available states are
        enumerated in a ``BidirectionalEnum`` at ``self.State``. Current state
        is stored in the object-private ``__state__``.

        :param state: ``State`` to set the current socket to.
        :returns: ``self``, for chainability. """

    return setattr(self, '__state__', state) or self

  def receive(self):

    """ Callable to reach through to the currently-active runtime and attempt
        to receive communication from the client-side of an active realtime
        communication session.

        :returns: Next message from the client (in a blocking/nonblocking way
         depending on the runtime's implementation). """

    return self.__runtime__.receive()

  def send(self, payload):

    """ Send a payload, via the currently-active runtime, to the client-side of
        the current realtime communication session.

        :param payload: Package of data to send to the client via the currently
          active realtime communication session.

        :returns: ``None``. """

    self.__runtime__.send(buffer(payload) if (
      isinstance(payload, bytearray)) else payload,
                          binary=isinstance(payload, bytearray))

  def close(self, graceful=True):

    """ Close the current realtime communication session (``graceful``ly or not,
        at the end of this function, it shall be closed).

        :param graceful: Attempt to close the connection with a ``FIN`` packet
          instead of simply resetting it.

        :returns: ``self``, for chainability. """

    self.__runtime__.close(graceful)
    return self

  # ~~ accessors ~~ #
  id, state, local, remote, runtime, established = (
    property(lambda self: self.__id__),
    property(lambda self: self.__state__),
    property(lambda self: self.__local__),
    property(lambda self: self.__remote__),
    property(lambda self: self.__runtime__),
    property(lambda self: self.__established__))


@decorators.bind('realtime', namespace=True)
class RealtimeSemantics(logic.Logic):

  """ Provides logic structure to handle realtime-style (acyclic) dispatch
      flows, as opposed to HTTP-style (cyclic) ones. Used in cases like
      WebSockets that provide a live bidirectional link. """

  hint = _SOCKET_KEY_ENV_ITEM

  def stream(self, target, send):  # pragma: no cover

    """ Prepare to consume a stream of packets/messages from ``target``, using
        ``send`` as a callback to transmit messages back to the client.

        Returns a ``responder`` generator that acts as a coroutine-style Python
        generator. Consumes messages via ``send`` and ``yield``s messages to be
        sent.

        :param target: Callable for consuming/receiving/blocking on new data
          from the client.

        :param send: Callable for sending data to the client.

        :returns: Prepared ``responder`` generator. """

    def responder(dispatch):

      """ Inner closure ``responder``, acts as a coroutine-style Python
          generator to consume packets from a realtime-enabled client
          connection, ``dispatch`` for a potential response, and forward
          outbound messages back to the client.

          :param dispatch: Application callable that should handle the inbound
            message, potentially ``yield``ing messages to be sent.

          :returns: Never. """

      while True:
        inbound = (yield)
        if inbound is TERMINATE: raise TerminateSocket(graceful=True)
        if inbound is not None:
          for outbound in dispatch(inbound):
            if outbound is TERMINATE: raise TerminateSocket(graceful=True)
            send(outbound)

    gen = responder(target)
    gen.next()
    return gen

  def on_connect(self, handler):  # pragma: no cover

    """ Top-level ``on_connect`` logic that provides initialization for a
        ``RealtimeSocket`` session.

        :param handler: Active ``RealtimeHandler`` instance that is handling
          this realtime session.

        :returns: :py:class:`RealtimeSocket` instance describing the
          newly-active realtime session. """

    # perform handshake through current runtime
    handler.runtime.handshake(*(
      handler.request.environ[self.hint],
      handler.request.environ.get(_ORIGIN_ENV_ITEM, '')))

    # resolve proxies
    sock = RealtimeSocket(handler.runtime, **{
      'local': handler.request.host,
      'remote': handler.request.remote_addr})

    sock.set_state(sock.State.INIT)  # set to init state initially
    handler.on_connect()  # provide to handler hook
    sock.set_state(sock.State.OPEN)  # set to open state after `on_connect`
    return sock

  def on_message(self, handler, socket):  # pragma: no cover

    """ Top-level ``on_message` logic that provides dispatch for inbound
        realtime messages.

        :param handler: Active ``RealtimeHandler`` instance that is handling
          this realtime session.

        :param socket: Active ``RealtimeSocket`` instance for the currently-
          active realtime session.

        :returns: ``None``, upon connection close. """

    responder = self.stream(handler.on_message, socket.send)
    socket.set_state(socket.State.ACTIVE)

    _terminate = False  # terminate flag
    try:
      while not _terminate:
        try:
          inbound = socket.receive()
          if inbound == TERMINATE: raise TerminateSocket(graceful=True)
          elif inbound: responder.send(inbound)

        except TerminateSocket as e:
          if e.graceful: _terminate = True  # terminate on re-loop
          else: raise  # re-raise and treat as an error

        except (GeneratorExit, StopIteration):
          break

    except TerminateSocket:
      # no way to gracefully terminate so
      pass

    finally:
      # was not gracefully terminated
      if not _terminate:  socket.set_state(socket.State.ERROR)
      self.on_close(handler, socket)

  def on_close(self, handler, socket):  # pragma: no cover

    """ Top-level ``on_close`` logic that provides cleanup logic for realtime
        communication sessions.

        :param handler: Currently-active ``RealtimeHandler`` instance that is
          managing this realtime session.

        :param socket: ``RealtimeSocket`` object describing the currently-
          active realtime session.

        :returns: ``self``, for chainability. """

    if not socket.state is socket.State.ERROR:
      socket.set_state(socket.State.CLOSED)
    handler.on_close(socket.state is not socket.State.ERROR)
    return self

__all__ = (
  'RealtimeSocket',
  'RealtimeSemantics')



# -*- coding: utf-8 -*-

"""

  struct tests
  ~~~~~~~~~~~~

  tests for canteen's data structures utilities.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# testing
from canteen import test

# utils
from canteen.util import struct


class BaseUtilTests(test.FrameworkTest):

  """ Tests toplevel stuff on :py:mod:`canteen.util`. """

  def test_exports(self):

    """ Test that `canteen.util` exports expected things """

    assert hasattr(struct, 'EMPTY')
    assert hasattr(struct, '_TOMBSTONE')


class SentinelTests(test.FrameworkTest):

  """ Tests :py:class:`canteen.util.struct.Sentinel`, which represents a simple
      singleton sentinel value.

      Sentinels are kind of like The Highlander... """

  def test_existence(self):

    """ Test basic existence of `util.struct.Sentinel` """

    assert hasattr(struct, 'Sentinel')

  def test_construct(self):

    """ Test basic functionality of `util.struct.Sentinel` """

    SAMPLE = struct.Sentinel('SAMPLE')
    assert SAMPLE.name == "SAMPLE"

  def test_equality(self):

    """ Test basic equality comparison for `util.struct.Sentinel` """

    SAMPLE = struct.Sentinel('SAMPLE')
    assert SAMPLE.name == "SAMPLE"

    SAMPLE2 = struct.Sentinel('SAMPLE')
    assert SAMPLE == SAMPLE2

  def test_repr(self):

    """ Test basic string representation of a `util.struct.Sentinel` """

    SAMPLE = struct.Sentinel('SAMPLE')
    assert 'SAMPLE' in str(SAMPLE)

  def test_falsy(self):

    """ Test ability to set a Sentinel as falsy. """

    BAD = struct.Sentinel('BAD', falsy=True)
    assert not BAD

  def test_not_falsy(self):

    """ Test ability to set a Sentinel as truthy. """

    GOOD = struct.Sentinel('GOOD')
    assert GOOD


class UtilStructTests(test.FrameworkTest):

  """ Tests :py:class:`util.struct.UtilStruct`, which is used as a base class
      for utility data structures. """

  def test_construct(self):

    """ Test that `UtilStruct` is abstract """

    with self.assertRaises(NotImplementedError):
      struct.UtilStruct()

  def test_fillstruct_abstract(self):

    """ Test that `UtilStruct.fillStructure is abstract """

    with self.assertRaises(TypeError):


      class UtilStructBadImplementor(struct.UtilStruct):

        """ Bad implementor of `UtilStruct` that should always raise a
            `TypeError` upon instantiation. """

        def i_am_not_fill_struct(self):

          """ I am not `fillStruct`. """

          return False  # pragma: no cover

      UtilStructBadImplementor()

    with self.assertRaises(NotImplementedError):

      class UtilStructBadSuper(struct.UtilStruct):

        """ Bad implementor of `UtilStruct` that should always raise a
            `NotImplementedError` because of invalid super access. """

        def fillStructure(self, _struct, case_sensitive=False, **kwargs):

          """ I am an invalid `fillStruct`. """

          _s = super(UtilStructBadSuper, self)
          _s.fillStructure(*(_struct, case_sensitive), **kwargs)

      UtilStructBadSuper().fillStructure({'blab': 'blab'})


class ObjectProxyTests(test.FrameworkTest):

  """ Tests :py:class:`util.struct.ObjectProxy`, which makes a ``dict``-like
      object usable via attribute syntax. """

  def test_construct(self):

    """ Test that `util.ObjectProxy` can be constructed. """

    # basic construction test
    st = struct.ObjectProxy()

    # construction test with struct
    st_struct = struct.ObjectProxy({
      'hi': True,
      'iam': False,
      'astruct': None})

    # construction test with kwargs
    st_kwargs = struct.ObjectProxy(hi=1, iam=2, astruct=3)

    # construction test with struct + kwargs
    st_both = struct.ObjectProxy({
      'hi': True,
      'iam': False,
      'astruct': None}, hi=1, iam=2)

  def test_fill_case_sensitive(self):

    """ Test that `util.ObjectProxy` can be case sensitive. """

    st_struct = struct.ObjectProxy({
      'HelloThere': True,
      'IamA': False,
      'StRuct': None}, case_sensitive=True)

    assert st_struct.HelloThere is True
    assert not hasattr(st_struct, 'hellothere')
    assert 'HelloThere' in st_struct
    assert 'hellothere' not in st_struct
    assert 'idonotexist' not in st_struct
    assert 'IDoNotExist' not in st_struct

  def test_fill_case_insensitive(self):

    """ Test that `util.ObjectProxy` can be case insensitive. """

    st_struct = struct.ObjectProxy({
      'HelloThere': True,
      'IamA': False,
      'StRuct': None}, case_sensitive=False)

    assert st_struct.HelloThere is True
    assert st_struct.hellothere is True
    assert 'HelloThere' in st_struct
    assert 'hellothere' in st_struct
    assert 'idonotexist' not in st_struct

  def test_getitem(self):

    """ Test that `util.ObjectProxy` can be used with getitem syntax. """

    st_struct = struct.ObjectProxy({
      'HelloThere': True,
      'IamA': False,
      'StRuct': None}, case_sensitive=False)

    assert st_struct['HelloThere'] is True
    assert st_struct['hellothere'] is True
    assert 'HelloThere' in st_struct
    assert 'hellothere' in st_struct
    assert 'idonotexist' not in st_struct

    # try invalid names
    with self.assertRaises(KeyError):
      st_struct['IDoNotExist']
    with self.assertRaises(KeyError):
      st_struct['idonotexist']

  def test_getattr(self):

    """ Test that `util.ObjectProxy` can be used with getitem syntax. """

    st_struct = struct.ObjectProxy({
      'HelloThere': True,
      'IamA': False,
      'StRuct': None}, case_sensitive=False)

    assert st_struct.HelloThere is True
    assert st_struct.hellothere is True
    assert 'HelloThere' in st_struct
    assert 'hellothere' in st_struct
    assert 'idonotexist' not in st_struct

    # try invalid names
    with self.assertRaises(AttributeError):
      st_struct.idonotexist
    with self.assertRaises(AttributeError):
      st_struct.IDoNotExist

    # test special attributes
    assert st_struct._case_sensitive is False
    assert isinstance(st_struct._entries, dict)

  def test_keys(self):

    """ Test buffered iteration with `util.ObjectProxy.keys`. """

    st_struct = struct.ObjectProxy({
      'HelloThere': True,
      'IamA': False,
      'StRuct': None}, case_sensitive=True)

    ref = ('HelloThere', 'IamA', 'StRuct')
    for key in st_struct.keys():
      assert key in ref

  def test_iterkeys(self):

    """ Test streaming iteration with `util.ObjectProxy.iterkeys`. """

    st_struct = struct.ObjectProxy({
      'HelloThere': True,
      'IamA': False,
      'StRuct': None}, case_sensitive=True)

    ref = ('HelloThere', 'IamA', 'StRuct')
    for key in st_struct.iterkeys():
      assert key in ref

    assert isinstance(st_struct.iterkeys(), {}.iterkeys().__class__)

  def test_values(self):

    """ Test buffered iteration with `util.ObjectProxy.values`. """

    st_struct = struct.ObjectProxy({
      'HelloThere': 1,
      'IamA': 2,
      'StRuct': 3})

    ref = (1, 2, 3)
    for val in st_struct.values():
      assert val in ref

  def test_itervalues(self):

    """ Test streaming iteration with `util.ObjectProxy.itervalues`. """

    st_struct = struct.ObjectProxy({
      'HelloThere': 1,
      'IamA': 2,
      'StRuct': 3})

    ref = (1, 2, 3)
    for val in st_struct.itervalues():
      assert val in ref

    assert isinstance(st_struct.itervalues(), {}.itervalues().__class__)

  def test_items(self):

    """ Test buffered iteration with `util.ObjectProxy.items`. """

    st_struct = struct.ObjectProxy({
      'HelloThere': True,
      'IamA': False,
      'StRuct': None}, case_sensitive=True)

    ref = {'HelloThere': True, 'IamA': False, 'StRuct': None}
    for key, value in st_struct.items():
      assert ref[key] == value

  def test_iteritems(self):

    """ Test streaming iteration with `util.ObjectProxy.iteritems`. """

    st_struct = struct.ObjectProxy({
      'HelloThere': True,
      'IamA': False,
      'StRuct': None}, case_sensitive=True)

    ref = {'HelloThere': True, 'IamA': False, 'StRuct': None}
    for key, value in st_struct.iteritems():
      assert ref[key] == value

    assert isinstance(st_struct.iteritems(), {}.iteritems().__class__)


class WritableObjectProxyTests(test.FrameworkTest):

  """ Tests :py:class:`util.struct.WritableObjectProxy`,
      which is like :py:class:`util.struct.ObjectProxy`
      but allows writes at runtime. """

  def test_setitem(self):

    """ Test that `util.WritableObjectProxy` can be used with setitem syntax """

    st_struct = struct.WritableObjectProxy()
    st_struct['hi'] = True

    assert 'hi' in st_struct
    assert st_struct['hi'] is True

  def test_setattr(self):

    """ Test that `util.WritableObjectProxy` can be used with setattr syntax """

    st_struct = struct.WritableObjectProxy()
    st_struct.hi = True

    assert 'hi' in st_struct
    assert st_struct.hi is True

  def test_delitem(self):

    """ Test that `util.WritableObjectProxy` can be used with delitem syntax """

    st_struct = struct.WritableObjectProxy()
    st_struct['hi'] = True
    st_struct['bye'] = False

    assert st_struct['hi'] is True
    assert st_struct['bye'] is False

    del st_struct['bye']

    assert 'bye' not in st_struct
    with self.assertRaises(KeyError):
      st_struct['bye']

    # try deleting an invalid item
    with self.assertRaises(KeyError):
      del st_struct['i_was_never_here_lol']

  def test_delattr(self):

    """ Test that `util.WritableObjectProxy` can be used with delattr syntax """

    st_struct = struct.WritableObjectProxy()
    st_struct.hi = True
    st_struct.bye = False

    assert st_struct.hi is True
    assert st_struct.bye is False

    del st_struct.bye

    assert 'bye' not in st_struct
    with self.assertRaises(AttributeError):
      st_struct.bye

    # try deleting an invalid attr
    with self.assertRaises(AttributeError):
      del st_struct.i_was_never_here_lol


class CallbackProxy(test.FrameworkTest):

  """ Tests :py:class:`util.struct.CallbackProxy`, which dispatches a callable
      to satisfy calls via entries based either on a known set of keys or a
      direct call to the callable with the desired attribute. """

  def test_construct_keys(self):

    """ Test constructing a `CallbackProxy` with registered keys """

    addone = lambda key: key + 1
    _struct = {
        'one': 0,
        'two': 1}

    proxy = struct.CallbackProxy(addone, _struct)

    # proxies with structs can be used with `in`
    assert 'one' in proxy
    assert 'two' in proxy
    assert proxy['one'] is 1
    assert proxy['two'] is 2

    return proxy

  def test_construct_keys_merge(self):

    """ Test merging registered keys passed to `CallbackProxy` """

    addone = lambda key: key + 1
    _struct = {
        'one': 0,
        'two': 1}

    proxy = struct.CallbackProxy(addone, _struct, two=4, three=3)

    assert proxy['one'] == 1
    assert proxy['two'] == 5  # because we overrode it
    assert proxy['three'] == 4  # because we added is

  def test_construct_nokeys(self):

    """ Test constructing a `CallbackProxy` with unregistered keys """

    addhi = lambda attr: 'hi' + attr
    proxy = struct.CallbackProxy(addhi)

    assert proxy['one'] == 'hione'
    assert proxy['two'] == 'hitwo'

    return proxy

  def test_valid_getattr_with_keys(self):

    """ Test getting a valid attr from a `CallbackProxy` with keys """

    proxy = self.test_construct_keys()
    assert proxy.one == 1
    assert proxy.two == 2

  def test_invalid_getattr_with_keys(self):

    """ Test getting an invalid attr from a `CallbackProxy` with keys """

    proxy = self.test_construct_keys()
    assert proxy.one == 1
    assert proxy.two == 2
    with self.assertRaises(AttributeError):
      assert proxy.three

  def test_valid_getattr_with_nokeys(self):

    """ Test getting a valid attr from a `CallbackProxy` with no keys """

    proxy = self.test_construct_nokeys()
    assert proxy.one == 'hione'
    assert proxy.two == 'hitwo'
    assert proxy.blab == 'hiblab'

  def test_invalid_getattr_with_nokeys(self):

    """ Test getting an invalid attr from a `CallbackProxy` with no keys """

    def _callback(request):

      """ Conditionally raise an ``AttributeError``. """

      if request is 'woops':
        raise AttributeError('Don\'t ask for woops')
      return 'blab' + request

    proxy = struct.CallbackProxy(_callback)
    assert proxy.one == 'blabone'
    assert proxy.two == 'blabtwo'
    assert proxy.blab == ('blab' * 2)

    with self.assertRaises(AttributeError):
      assert proxy.woops

  def test_valid_getitem_with_keys(self):

    """ Test getting a valid item from a `CallbackProxy` with keys """

    proxy = self.test_construct_keys()
    assert proxy['one'] == 1
    assert proxy['two'] == 2

  def test_invalid_getitem_with_keys(self):

    """ Test getting an invalid item from a `CallbackProxy` with keys """

    proxy = self.test_construct_keys()
    assert proxy['one'] == 1
    assert proxy['two'] == 2
    with self.assertRaises(KeyError):
      assert proxy['three']

  def test_valid_getitem_with_nokeys(self):

    """ Test getting a valid item from a `CallbackProxy` with no keys """

    proxy = self.test_construct_nokeys()
    assert proxy['one'] == 'hione'
    assert proxy['two'] == 'hitwo'
    assert proxy['blab'] == 'hiblab'

  def test_invalid_getitem_with_nokeys(self):

    """ Test getting an invalid item from a `CallbackProxy` with no keys """

    def _callback(request):

      """ Conditionally raise an ``KeyError``. """

      if request is 'woops':
        raise KeyError('Don\'t ask for woops')
      return 'blab' + request

    proxy = struct.CallbackProxy(_callback)
    assert proxy['one'] == 'blabone'
    assert proxy['two'] == 'blabtwo'
    assert proxy['blab'] == ('blab' * 2)

    with self.assertRaises(KeyError):
      assert proxy['woops']


class BidirectionalEnumTests(test.FrameworkTest):

  """ Tests the utility structure :py:class:`BidirectionalEnum`, which makes
      enumeration structures that map keys to values and values to keys. """

  def test_construct(self):

    """ Test constructing a `BidirectionalEnum` """

    class Colors(struct.BidirectionalEnum):

      """ Enumerates pretty colors. """

      BLUE = 0x0
      RED = 0x1
      GREEN = 0x2

    return Colors

  def test_abstract(self):

    """ Test abstractness of `BidirectionalEnum` """

    with self.assertRaises(TypeError):
      struct.BidirectionalEnum()

  def test_contains(self):

    """ Test `x in y` syntax against `BidirectionalEnum` """

    enum = self.test_construct()
    assert 'BLUE' in enum
    assert 'RED' in enum
    assert 'GREEN' in enum
    assert 'BLACK' not in enum
    assert 'GRAY' not in enum

    # should work in reverse too
    assert 0x0 in enum
    assert 0x1 in enum
    assert 0x2 in enum
    assert 0x3 not in enum
    assert 0x4 not in enum

  def test_getattr(self):

    """ Test `x.y` syntax against `BidirectionalEnum` """

    enum = self.test_construct()
    assert enum.BLUE is enum.BLUE is 0x0
    assert enum.RED is enum.RED is 0x1
    assert enum.GREEN is enum.GREEN is 0x2

    with self.assertRaises(AttributeError):
      assert enum.BLACK

  def test_getitem(self):

    """ Test `x[y]` syntax against `BidirectionalEnum` """

    enum = self.test_construct()
    assert enum['BLUE'] is enum['BLUE'] is 0x0
    assert enum['RED'] is enum['RED'] is 0x1
    assert enum['GREEN'] is enum['GREEN'] is 0x2

    # reverse-resolve fallback should be supported
    assert enum[0x0] == 'BLUE'
    assert enum[0x1] == 'RED'
    assert enum[0x2] == 'GREEN'

    with self.assertRaises(KeyError):
      assert enum['BLACK']

  def test_immutable(self):

    """ Test `BidirectionalEnum` for immutability """

    enum = self.test_construct()

    # disallow new properties
    with self.assertRaises(NotImplementedError):
     enum['BLACK'] = 0x5

    with self.assertRaises(NotImplementedError):
      enum.BLACK = 0x5

    # disallow overwrites too
    with self.assertRaises(NotImplementedError):
      enum['GREEN'] = 0x5

    with self.assertRaises(NotImplementedError):
      enum.GREEN = 0x5

    # make sure nothing changed
    assert enum['GREEN'] is enum.GREEN is 0x2

  def test_reverse_resolve(self):

    """ Test reverse-resolving a key from a value with a `BidirectionalEnum` """

    enum = self.test_construct()

    assert enum.reverse_resolve(0x0) == 'BLUE'
    assert enum.reverse_resolve(0x1) == 'RED'
    assert enum.reverse_resolve(0x2) == 'GREEN'

    # reverse-resolves fail softly
    assert not enum.reverse_resolve(0x3)
    assert not enum.reverse_resolve(0x3)

  def test_iter(self):

    """ Test iteration against `BidirectionalEnum` """

    enum = self.test_construct()

    _struct = {
        'BLUE': 0x0,
        'RED': 0x1,
        'GREEN': 0x2}

    for key, value in enum:
      assert key in _struct
      assert _struct[key] is value
      _struct[key] = True

    assert all(_struct.itervalues())  # make sure all values touched



# -*- coding: utf-8 -*-

"""

  model adapters
  ~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# module constants
__version__ = (0, 8)  # module version-string
__doc__ = "Contains modules that adapt models to various storage backends."


# abstract adapters
from . import abstract
from .abstract import Mixin
from .abstract import KeyMixin
from .abstract import ModelMixin
from .abstract import VertexMixin
from .abstract import EdgeMixin
from .abstract import ModelAdapter
from .abstract import IndexedModelAdapter

abstract_adapters = [abstract, ModelAdapter, IndexedModelAdapter]


# adapter modules
from . import redis
from . import inmemory
from . import protorpc

modules = [inmemory, protorpc, redis]


# concrete adapters
from .redis import RedisAdapter
from .inmemory import InMemoryAdapter

concrete = [InMemoryAdapter, RedisAdapter]


# builtin mixins
from . import core
from .core import DictMixin
from .core import JSONMixin

base_mixins = [KeyMixin, ModelMixin, VertexMixin, EdgeMixin]
builtin_mixins = [DictMixin, JSONMixin]


__adapters__ = tuple((
  abstract_adapters +
  modules +
  concrete +
  builtin_mixins +
  base_mixins))



# -*- coding: utf-8 -*-

"""

  core meta
  ~~~~~~~~~

  metaclass tools and APIs.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# utils
from ..util import struct, decorators


## Globals
_owner_map = {}
grab = lambda x: x.__func__ if hasattr(x, '__func__') else x
owner = lambda x: intern(x.__owner__ if hasattr(x, '__owner__') else x.__name__)
construct = lambda cls, n, b, p: type.__new__(cls, n, b, p)
metachain = lambda cls, n, b, p: (
  cls.__metachain__.append(construct(cls, n, b, p)) or cls.__metachain__[-1])
is_component = lambda x: issubclass(x.__class__, Proxy.Component)


class MetaFactory(type):

  """ Meta class factory, used to prepare core metaclasses for use as
      functionality sentinels that compound functionality all the way up the
      tree as they are defined.

      This class is meta-abstract, meaning it must be used as a parent or
      metaclass to a more concrete implementation and cannot be instantiated or
      used directly. """

  __owner__, __metachain__, __root__ = "BaseMeta", [], True

  def __new__(mcs, name=None, bases=None, properties=None):

    """ Construct a new ``MetaFactory`` concrete class, implementing the
        ``initialize`` protocol for bootstrapping meta-implementing classes
        dynamically, potentially with DI-enabled MRO.

        :param name: String name for a new metaimplementor of ``MetaFactory``.

        :param bases: Tuple of base classes in the calling definition.

        :param properties: ``dict`` map of ``key => value`` class-level
          attributes.

        :raises NotImplementedError: If direct construction is attempted (i.e.
          not via Python's metaclass internals or a manual call of the same
          signautre), as ``MetaFactory`` is abstract.

        :returns: Factoried ``MetaFactory`` meta-implementor. """

    if not name or not bases or not (
      isinstance(properties, dict)):  # pragma: no cover
      raise NotImplementedError('`MetaFactory` is meta-abstract'
                                ' and cannot be constructed directly.')

    # get ready to construct, do so immediately for ``MetaFactory`` itself
    if '__root__' in properties and properties['__root__']:
      del properties['__root__']  # treat as a root - init directly and continue
      return construct(mcs, name, bases, properties)

    # patch class name to owner if it's embedded
    name = name if (
        (name != '__metaclass__' or '__owner__' not in properties)) else (
          properties['__owner__'])

    # construct, yo. then unconditionally apply it to the meta chain and return
    # also, defer to the class' ``initialize``, or any of its bases if they have
    # ``initialize`, for constructing the actual class.
    return ((grab(properties['initialize'] if 'initialize' in properties else
                  getattr((x for x in bases if hasattr(x, 'initialize')).next(),
                          'initialize')))(*(mcs, name, bases, properties))) if (
                            'initialize' in properties or any((
                              hasattr(b, 'initialize') for b in bases))
                                ) else metachain(mcs, name, bases, properties)

  def mro(cls):

    """ Assemble MRO (Method Resolution Order) to enable proper class composure
        patterns for ``MetaFactory``.

        :returns: All available mixin classes are mixed into the target MRO to
          enable a fully composure-based inheritance system. """

    # override metaclass MRO to make them superimpose-able on each other
    if not cls.__metachain__:
      return type.mro(cls)

    # make sure to enforce MRO semantics
    seen, tree, order = set(), [], type.mro(cls)
    for group in ([order[0]], order[1:-2], cls.__metachain__, order[-2:]):
      for base in group:
        if base not in seen: seen.add(base), tree.append(base)
    return tuple(tree)

  __repr__ = lambda cls: "<meta '%s.%s'>" % (cls.__owner__, cls.__name__)


class Base(type):

  """ Acts as a concrete anchor to core metaclasses. Can be used in an
      ``isinstance`` check to identify classes used in the meta-construction and
      initialization of downstream types and objects. """

  __owner__, __metaclass__, __root__ = "Base", MetaFactory, True


class Proxy(object):

  """ Container class for core metaclasses. Used to package references to
      structures that radically change Python's class system, such that one must
      explicitly use a ``Proxy.something`` qualified path as a metaclass. """


  class Factory(Base):

    """ Metaclass that enforces a pattern whereby concrete classes are passed
        through a factory function for construction. """

    __hooks__ = []

    def initialize(cls, name, bases, properties):

      """ Construct a new ``Factory`` concrete class. Dispatched when
          ``Factory`` or further-downstream core structures are used as
          metaclasses.

          :param name: Name of the target class to factory.
          :param bases: Tuple of target class bases.
          :param properties: Map of class-level attributes.

          :returns: Prepared ``Factory`` initializor class. """

      def metanew(_cls, _name, _bases, _properties):

        """ Closure that overrides ``__new__`` to inject custom class
            construction behavior.

            :param _cls: Target class to inject custom construction behavior on.

            :param _name: Name of the downstream class to factory with custom
              construction behavior.

            :param _bases: Tuple of bases for the downstream class to factory
              with custom construction behavior.

            :param _properties: Map of class-level attributes for downstream
              class desiring custom construction behavior.

            :returns: Properly constructed (and potentially, registered via
              ``Proxy.Registry``) downstream target class. """

        # if this metaclass implements the ``Proxy.Register`` class,
        #  defer to _cls.register directly after construction
        if issubclass(_cls, Proxy.Registry):
          return grab(_cls.register)(_cls, construct(*(
            _cls, _name, _bases, _properties)))
        return construct(_cls, _name, _bases, _properties)  # pragma: nocover

      # drop down if we already have a metachain for this tree
      if cls.__metachain__: properties['__new__'] = metanew

      # construct, yo. then unconditionally apply it to the metachain and return
      return metachain(cls, name, bases, properties)


  class Registry(Factory):

    """ Metaclass that enforces a pattern whereby classes are registered in a
        central datastructure according to their ``__bases__`` directly before
        or after class construction. """

    __chain__ = {}

    def iter_children(cls):

      """ Iterate over a parent class' registered child classes, one at a time.

          :returns: Yields child classes, one at a time. """

      for obj in cls.__chain__[owner(cls)]:
        if obj is cls: continue  # skip the parent class
        yield obj

    def children(cls):

      """ Retrieve a list of all this parent class' registered children.

          :returns: ``list`` of this ``cls``'s children. """

      # remember to filter-out weakrefs that have died
      return [child for child in cls.iter_children()]

    @staticmethod
    def register(meta, target):

      """ Register a new constructed subclass at ``target``, utilizing
          ``meta``'s chain.

          :param meta: Base ``chain`` that should be used for to register
            the ``target`` meta subclass.

          :param target: Target meta subclass to add to ``meta``'s metachain
            and register.

          :returns: ``target`` that was registered. """

      _owner = owner(target)

      # check to see if bases are only roots, if it is a root
      # create a new metabucket
      if not any((
        (False if x in (object, type) else True) for x in target.__bases__)):
        meta.__chain__[_owner] = []
        return target

      # resolve owner and construct
      for base in target.__bases__:
        if not base in (object, type):
          if _owner not in meta.__chain__: meta.__chain__[_owner] = []
          meta.__chain__[_owner].append(target)
      return target


  class Component(Registry):

    """ Decorate a class tree as capable of being injected as DI ``components``,
        which are bound to simple string names and made available
        application-wide at ``self``. """

    __target__ = None
    __binding__ = None
    __injector_cache__ = {}
    __map__ = {}  # holds map of all platform instances

    @decorators.classproperty
    def singleton_map(cls):

      """ Retrieve the application-wide map of singleton classes, bound to their
          names.

          :returns: Map of known singletons for the current execution flow. """

      return cls.__map__

    @classmethod
    def reset_cache(mcs):

      """ Reset injector caches. """

      mcs.__injector_cache__ = {}
      mcs.__class__.__injector_cache__ = {}

    @classmethod
    def prepare(mcs, target):

      """ Prepare ``target`` (usually ``cls``) for injection, possibly resolving
          a global singleton object to be returned upon matching attribute
          requests.

          :param target: Target class to prepare for later injection.

          :returns: Bound ``target``, after mapping any bound ``__binding__``
            aliases, methods, etc. """

      if (not hasattr(target, '__binding__')) or target.__binding__ is None:
        return  # non-bound classes don't need preparation

      alias = (  # resolve name, instantiate and register instance singleton
        target.__binding__.__alias__ if (
          hasattr(target.__binding__, '__alias__') and (
            isinstance(target.__binding__, basestring))) else target.__name__)

      if hasattr(target, '__singleton__') and target.__singleton__:
        # if we already have a singleton, give that
        if alias in mcs.__map__: return mcs.__map__[alias]

        # otherwise, startup a new singleton
        mcs.__map__[alias] = target()
        return mcs.__map__[alias]
      return target  # pragma: nocover

    @staticmethod
    def inject(cls):

      """ Parse/consider bindings attached to the target ``cls``, providing the
          final concrete class and set of injectable items.

          :returns: Two-item tuple of 1) the final concrete class (after
            ``prepare`` is invoked) and a manifest of inejctable bindings,
              passed as a Python ``set``. """

      # allow class to "prepare" itself (potentially instantiating a singleton)
      concrete = (cls.__class__.prepare(cls) if (
        hasattr(cls.__class__, 'prepare')) else cls)

      # allow class to indicate it does not wish to inject
      if concrete is None: return

      # gather injectable attributes
      _injectable = set()
      if hasattr(cls, '__bindings__'):
        for iterator in (cls.__dict__.iteritems(),
                         cls.__class__.__dict__.iteritems()):
          for prop, value in iterator:
            if cls.__bindings__:
              if prop in cls.__bindings__:
                func = (
                  cls.__dict__[prop] if not isinstance(cls.__dict__[prop], (
                    staticmethod, classmethod))
                  else cls.__dict__[prop].__func__)
                do_namespace = func.__binding__.__namespace__ if (
                  cls.__binding__.__namespace__) else False
                _injectable.add((
                  prop, func.__binding__.__alias__ or prop, do_namespace))

      # return bound injectables or the whole set
      return concrete, _injectable or (
        set(filter(lambda x: not x.startswith('__'),
            concrete.__dict__.iterkeys())))

    @staticmethod
    def collapse(cls, spec=None):

      """ Collapse available ``component`` items into a mapping of names to
          objects which can respond to attribute requests for those paths.

          :param spec: Base class to use as *perspective* for the DI collapse
            and response. Currently unused, but could easily be used to produce
            an aspect- or interface- based (stricter) DI system, rather than the
            wild-west-resolve-through-a-dictionary default in canteen.

          :returns: Collapsed DI pool, ready to be used for resolving dependency
            requests with registered attribute values at deeply-specified
            paths. """

      # try the injector cache
      if (cls, spec) not in Proxy.Component.__injector_cache__:

        # otherwise, collapse and build one
        property_bucket = {}
        for metabucket in Proxy.Registry.__chain__.iterkeys():
          for concrete in filter(is_component,
                                  Proxy.Component.__chain__[metabucket]):

            namespace = ''
            responder, properties = concrete.inject(concrete) or (None, {})

            # filter out classes that opt-out of injection
            if not responder: continue

            if hasattr(concrete, '__binding__'):

              # @TODO(sgammon): what the fuck does this shit even do
              def do_pluck(klass, obj, pool):

                """ First-level closure that prepares a ``pluck`` function to
                    properly grab an ``obj`` from the DI ``pool``, in the
                    context of ``klass``.

                    :param klass: Class context for which we would like to
                      prepare our ``pluck`` function. Exposes the base type to
                      the inner closure and keeps it around.

                    :param obj: Object singleton that can be used as a base for
                      resolving dependencies mapped to this class.

                    :param pool: Reference upwards to the main DI resource pool.

                    :returns: Prepared and closured ``pluck`` function, which is
                      particularly useful in a structure like a
                      :py:class:`struct.CallbackProxy`."""

                def pluck(property_name):

                  """ Second-level closure that plucks a property (at
                      ``property_name``) from the DI pool encapsulated in the
                      outer closure, in the context of ``klass``.

                      :param property_name: Property name that we would like to
                        resolve from ``klass`` or ``obj``, bound from the outer
                        closure function.

                      :raises AttributeError: If the requested ``property_name``
                        could not be resolved, by virtue of the fact that this
                        function makes use of ``getattr`` directly.

                      :returns: The value at the given ``property_name``, if
                        any, or raises ``AttributeError``, per standard Python
                        attribute resolution protocol. """

                  # dereference property aliases
                  if hasattr(klass, '__aliases__') and (
                    property_name in klass.__aliases__):
                    return getattr(obj, klass.__aliases__[property_name])
                  if hasattr(obj, property_name):
                    return getattr(obj, property_name)  # pragma: nocover

                  if klass.__binding__ and klass.__binding__.__alias__:
                    parent_ref = '.'.join((
                      klass.__binding__.__alias__, property_name))
                    if parent_ref in pool:
                      return pool[parent_ref]

                return setattr(pluck, 'target', klass) or pluck

              if concrete.__binding__:
                property_bucket[concrete.__binding__.__alias__] = (
                  struct.CallbackProxy(do_pluck(*(
                    concrete,
                    responder,
                    property_bucket
                  ))))

                if concrete.__binding__.__namespace__:
                  namespace = concrete.__binding__.__alias__

            for bundle in properties:

              # clear vars
              prop, alias, _global = None, None, False

              if not isinstance(bundle, tuple):  # pragma: no cover
                _key = '.'.join((namespace, bundle)) if namespace else bundle
                property_bucket[_key] = (responder, bundle)
                continue

              prop, alias, _global = bundle
              _key = alias
              if _global:
                _key = '.'.join((namespace, alias)) if namespace else alias
                property_bucket[_key] = (responder, prop)
                continue
              property_bucket[_key] = (responder, prop)

        if not property_bucket: return {}  # if it's empty, don't cache

        # set in cache, unless empty
        Proxy.Component.__injector_cache__[(cls, spec)] = property_bucket

      # return from cache
      return Proxy.Component.__injector_cache__[(cls, spec)]



# -*- coding: utf-8 -*-

"""

  model key tests
  ~~~~~~~~~~~~~~~

  tests canteen's model keys.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# canteen keys
from canteen import model
from canteen.model import Key
from canteen.model import exceptions
from canteen.model import AbstractKey

# canteen tests
from canteen.test import FrameworkTest


## KeyTests
class KeyTests(FrameworkTest):

  """ Tests `model.Key` and `model.AbstractKey`. """

  def test_construct_key(self):

    """ Test constructing a `Key` manually """

    # test kinded empty key
    k = Key("TestKind")
    self.assertEqual(k.kind, "TestKind")

    # test kinded ID'd key
    k = Key("TestKind", "sample")
    self.assertEqual(k.kind, "TestKind")
    self.assertEqual(k.id, "sample")

    # test parented, kinded ID'd key
    pk = Key("TestParentKind", "parent")
    k = Key("TestKind", "child", parent=pk)
    self.assertEqual(k.kind, "TestKind")
    self.assertEqual(k.id, "child")
    self.assertEqual(k.parent, pk)

    # make sure basic properties are ok
    self.assertEqual(k.__slots__, set())
    self.assertEqual(k.__class__.__name__, "Key")
    self.assertEqual(k.__class__.__slots__, set())

  def test_key_inheritance(self):

    """ Test proper inheritance structure for `Key` """

    # test basic key inheritance
    k = Key("TestKind", "sample")

    # test basic inheritance
    self.assertIsInstance(k, Key)
    self.assertIsInstance(k, AbstractKey)
    self.assertIsInstance(k, object)

    # test class inheritance
    self.assertTrue(issubclass(Key, AbstractKey))
    self.assertTrue(AbstractKey in Key.__bases__)
    self.assertIsInstance(Key, AbstractKey.__metaclass__)

    # check `AbstractKey` inheritance
    self.assertTrue(type(k) == Key)
    self.assertTrue(issubclass(Key, object))

  def test_key_stringify(self):

    """ Test string representation of a `Key` object """

    # build and stringify key
    k = Key("SampleKind", "sample_id")
    x = str(k)

    # make sure the kind is somewhere
    self.assertTrue(("kind" in x))
    self.assertTrue(("SampleKind" in x))

    # make sure the ID is somewhere
    self.assertTrue(("id" in x))
    self.assertTrue(("sample_id" in x))

    # make sure the key class is somewhere
    self.assertTrue(('Key' in x))

  def test_key_class_stringify(self):

    """ Test the string representation of a `Key` class """

    # build and stringify key
    x = str(Key)

    # make sure the kind is somewhere
    self.assertTrue(("kind" in x))

    # make sure the ID is somewhere
    self.assertTrue(("id" in x))

    # make sure the key class is somewhere
    self.assertTrue(('Key' in x))

  def test_abstract_key(self):

    """ Test that `AbstractKey` works abstractly """

    # should not be able to instantiate `AbstractKey`
    with self.assertRaises(exceptions.AbstractConstructionFailure):
      AbstractKey()

    self.assertIsInstance(Key(), AbstractKey)

  def test_abstract_key_concrete(self):

    """ Test that `AbstractKey` works concretely """

    # sample `Key` subclass
    class SampleKey(AbstractKey):
      """ Tests subclasses of `AbstractKey`. """
      __schema__ = ('id', 'kind')

    # perform tests
    self.assertTrue(SampleKey("Sample", "id"))
    self.assertIsInstance(SampleKey("Sample", "id"), Key)
    self.assertTrue(hasattr(SampleKey, '__schema__'))
    self.assertEqual(len(SampleKey.__schema__), 2)

  def test_key_concrete(self):

    """ Test that `Key` works concretely """

    # sample `Key` subclass
    class SampleKey(Key):
      """ Tests subclasses of `Key`. """
      __schema__ = ('id', 'kind')

    # perform tests
    self.assertTrue(SampleKey("Sample", "id"))
    self.assertIsInstance(SampleKey("Sample", "id"), Key)
    self.assertTrue(hasattr(SampleKey, '__schema__'))
    self.assertEqual(len(SampleKey.__schema__), 2)

  def test_raw_key_format(self):

    """ Test constructing a key from a raw iterable """

    # sample key
    k = Key("Sample", "sample")

    # tupled raw
    self.assertEqual(Key(raw=k.flatten()), k)
    self.assertEqual(Key.from_raw(k.flatten()), k)

    # joined raw
    joined, struct = k.flatten(True)
    self.assertEqual(Key(raw=joined), k)
    self.assertEqual(Key.from_raw(joined), k)

  def test_urlsafe_key_format(self):

    """ Test constructing a key from its encoded form """

    # sample key
    k = Key("Sample", "sample")

    # urlsafe in & out
    self.assertEqual(Key(urlsafe=k.urlsafe()), k)
    self.assertEqual(Key.from_urlsafe(k.urlsafe()), k)

  def test_key_flatten(self):

    """ Test flattening a Key into a raw iterable """

    # sample key
    k = Key("Sample", "sample")

    # flatten in & out
    self.assertEqual(Key(raw=k.flatten()), k)
    self.assertEqual(Key.from_raw(k.flatten()), k)

  def test_key_nonzero(self):

    """ Test nonzero functionality in a key """

    # sample zero key and nonzero key
    k, nk = Key("Sample"), Key("Sample", "sample")

    # key with no ID should evaluate to falsy
    self.assertTrue(nk)
    self.assertTrue(not k)

  def test_key_len(self):

    """ Test length of a `Key`, which should be 0 for an incomplete key """

    k, nk = Key("Sample"), Key("Sample", "sample")

    # a key with no ID should evaluate to 0 via len()
    self.assertEqual(len(k), 0)
    self.assertEqual(len(nk), 1)

  def test_key_with_model_class_kind(self):

    """ Test making a `Key` via using a model class as the kind """

    ## KindedModel
    # Used to test using classes for kinds in `model.Key`.
    class KindedModel(model.Model):

      """ Sample for testing key creation from model classes. """

      string = basestring

    # make keys
    k1 = model.Key("KindedModel", "test_id")
    k2 = model.Key(KindedModel, "test_id")
    ko = model.Key(KindedModel)

    # test keys
    self.assertEqual(k1.kind, "KindedModel")
    self.assertEqual(k1.id, "test_id")
    self.assertEqual(k2.kind, k1.kind)
    self.assertEqual(ko.kind, k1.kind)
    self.assertEqual(k2.id, k2.id)

  def test_key_ancestry(self):

    """ Test `Key` ancestry mechanisms """

    # manufacture keys
    pk = model.Key("ParentKind", "parent_id")
    ck = model.Key("ChildKind", "child_id", parent=pk)
    gk = model.Key("GrandchildKind", "grandchild_id", parent=ck)
    ggk = model.Key("GreatGrandchildKind", "great_grandchild_id", parent=gk)

    # for each key, make sure parent is set
    self.assertEqual(pk.parent, None)
    self.assertEqual(ck.parent, pk)
    self.assertEqual(gk.parent, ck)
    self.assertEqual(ggk.parent, gk)

    # for each key, make sure ancestry works
    pk_ancestry = [i for i in pk.ancestry]
    ck_ancestry = [i for i in ck.ancestry]
    gk_ancestry = [i for i in gk.ancestry]
    ggk_ancestry = [i for i in ggk.ancestry]

    # test ancestry paths
    self.assertEqual(len(pk_ancestry), 1)
    self.assertEqual(len(ck_ancestry), 2)
    self.assertEqual(len(gk_ancestry), 3)
    self.assertEqual(len(ggk_ancestry), 4)

    # len of a key should always be 1 unless it has ancestry,
    # then it's the length of the ancestry chain
    self.assertEqual(len(pk), 1)
    self.assertEqual(len(ck), 2)
    self.assertEqual(len(gk), 3)
    self.assertEqual(len(ggk), 4)

    # ... however all keys should test nonzero-ness (all keys should be nonzero)
    for k in (pk, ck, gk, ggk):
      self.assertTrue(k)

  def test_key_with_overflowing_schema(self):

    """ Test construction of a `Key` with too many schema items """

    # try and make a key with a ton of arguments
    with self.assertRaises(TypeError):
      model.Key(*("SampleKind",
                  "id",
                  "coolstring",
                  "whatdowedo",
                  "whenwehave",
                  "thismanyarguments"))

  def test_key_construct_multiple_formats(self):

    """ Test constuction of a `Key` with multiple formats """

    # sample key
    ok = model.Key("Sample", "sample_id")

    # try and make a key with multiple formats
    with self.assertRaises(TypeError):
      model.Key(raw=ok.flatten(False)[1], urlsafe=ok.urlsafe())

  def test_key_auto_id(self):

    """ Test an integer-based ID field """

    class AutoincrementTest(model.Model):

      """ Test that keys autoincrement properly when not assigned
        deterministic name values. """

      message = basestring, {'default': 'Hello, world!'}

    # put deterministic key
    a = AutoincrementTest(key=model.Key(AutoincrementTest.kind(),
                          'testing-string-key'))
    dk = a.put()

    # put nondeterministic key #1
    nk = AutoincrementTest().put()

    # put nondeterministic key #2
    nk2 = AutoincrementTest().put()

    self.assertIsInstance(nk.id, int)  # nondeterministic should be an int
    self.assertIsInstance(dk.id, basestring)  # deterministic should be a string
    self.assertTrue((nk2.id > nk.id))  # should be greater

  def test_key_adapter(self):

    """ Test that the adapter is attached correctly to `Key` """

    # build test obj
    k = model.Key("TestKind", "test")

    # make sure an adapter is attached at the class level
    self.assertTrue(hasattr(model.Key, '__adapter__'))
    self.assertIsInstance(model.Key.__adapter__, model.adapter.ModelAdapter)

    # make sure the same is true at the object level
    self.assertTrue(hasattr(k, '__adapter__'))
    self.assertIsInstance(model.Key.__adapter__, model.adapter.ModelAdapter)

  def test_key_equality(self):

    """ Test that keys equal each other when they should """

    conditions = []

    # test with determininstic ID
    k1 = model.Key("TestKind", "blabs")
    k2 = model.Key("TestKind", "blabs")
    conditions.append((k1 == k2))

    # test kinded keys
    k3 = model.Key("TestKind")
    k4 = model.Key("TestKind")
    conditions.append((k3 == k4))

    # test ancestored keys
    k5 = model.Key("TestSubkind", "blobs", parent=k1)
    k6 = model.Key("TestSubkind", "blobs", parent=k2)
    conditions.append((k5 == k6))

    [self.assertTrue(condition) for condition in conditions]

  def test_key_inequality(self):

    """ Test that keys don't equal each other when they shouldn't """

    conditions = []

    # test with determininstic ID
    k1 = model.Key("TestKind", "blabs")
    k2 = model.Key("TestKind", "blobs")
    conditions.append((k1 != k2))

    # test kinded keys
    k3 = model.Key("TestKind")
    k4 = model.Key("TestOtherKind")
    conditions.append((k3 != k4))

    # test ancestored keys with different IDs
    k5 = model.Key("TestSubkind", "blabs", parent=k1)
    k6 = model.Key("TestSubkind", "blobs", parent=k1)
    conditions.append((k5 != k6))

    # test ancestored keys with different kinds
    k7 = model.Key("TestSubkindOne", "blabs", parent=k1)
    k8 = model.Key("TestSubkindTwo", "blabs", parent=k1)
    conditions.append((k7 != k8))

    # test ancestored keys with different parents
    k9 = model.Key("TestSubkind", "blabs", parent=k1)
    k10 = model.Key("TestSubkind", "blabs", parent=k2)
    conditions.append((k9 != k10))

    [self.assertTrue(condition) for condition in conditions]

  def test_key_format(self):

    """ Test format specification on `Key` """

    # build object
    k1 = model.Key("Test", "testkey")

    # test class
    self.assertTrue(hasattr(model.Key, '__schema__'))
    self.assertIsInstance(model.Key.__schema__, tuple)
    self.assertTrue((len(model.Key.__schema__) > 1))

    # test object
    self.assertTrue(hasattr(k1, '__schema__'))
    self.assertIsInstance(model.Key.__schema__, tuple)
    self.assertTrue((len(model.Key.__schema__) > 1))

  def test_key_set_unknown_attribute(self):

    """ Test setting an unknown and known attribute on a `Key` """

    k = model.Key("CoolKind", "coolid")
    with self.assertRaises(AttributeError):
      k.blabble = True  # try writing unknown attribute

  def test_key_overwrite_known_attribute(self):

    """ Test overwriting a known (schema-d) attribute on a `Key` """

    k = model.Key("CoolKind", "coolid")
    with self.assertRaises(AttributeError):
      k.kind = "MyKind"
    with self.assertRaises(AttributeError):
      k.id = 10

  def test_key_set_attribute_persisted(self):

    """ Test setting a valid attribute on a persisted `Key` """

    class PersistedKeyTest(model.Model):

      """ Test model for making sure writing to a persisted key fails. """

      message = basestring, {'default': 'Hello, world!'}

    with self.assertRaises(AttributeError):
      k = PersistedKeyTest().put()
      k.id = 5

    with self.assertRaises(AttributeError):
      k = PersistedKeyTest().put()
      k._set_internal('kind', "Blabs")

    with self.assertRaises(AttributeError):
      k = PersistedKeyTest().put()
      k._set_internal("id", 25)



# -*- coding: utf-8 -*-

"""

  base handler tests
  ~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# testing
from canteen import test

# base, core + logic
from canteen.base import handler
from canteen.logic import template
from canteen.logic.http import semantics
from canteen.core import runtime as rtime

# utils & template logic
from canteen.util.config import Config

# mock objects
from werkzeug import test as wtest
from werkzeug.wrappers import Request
from werkzeug.wrappers import Response
from werkzeug.exceptions import MethodNotAllowed


class BaseHandlerTest(test.FrameworkTest):

  """ Tests for :py:mod:`canteen.base.handler`. """

  def _make_handler(self, valid=False, _impl=handler.Handler, _environ=None):

    """ Mock up a quick `Handler` object. """

    if not valid:  # don't need valid WSGI internals?
      _environ = wtest.EnvironBuilder().get_environ()
      _environ['sample'] = 'hi'
      environ, callback, _runtime, request, response = (
        _environ, lambda: True, rtime.Runtime(object()),
        semantics.HTTPSemantics.new_request(_environ),
        semantics.HTTPSemantics.new_response())

      # allow environ overrides
      if _environ: environ.update(_environ)

      canteen_style = _impl(*(
        environ, callback, _runtime, request, response))
      return canteen_style, request, response, _runtime

    result, environ = (
      {},
      wtest.create_environ('/sample', 'http://localhost:8080/'))

    # allow environ overrides
    if _environ: environ.update(_environ)

    callback, _runtime, request, response = (
      lambda status, headers: (
        result.__setitem__('status', status) and
        result.__setitem__('headers', headers)),
      rtime.Runtime(object()),
      Request(environ),
      Response())

    canteen_style = _impl(*(
      environ,
      callback,
      _runtime,
      request,
      response))

    return canteen_style, request, response, _runtime

  def test_base_handler(self):

    """ Test that `Handler` is exposed for import """

    assert hasattr(handler, 'Handler')

  def test_construct_handler(self):

    """ Test various constructions of `Handler` """

    ## try WSGI-style construction
    environ = {'sample': 'hi'}
    callback = lambda: True

    wsgi_style = handler.Handler(environ, callback)
    assert hasattr(wsgi_style, 'environ')
    assert hasattr(wsgi_style, 'environment')
    assert wsgi_style.environ['sample'] == 'hi'
    assert wsgi_style.callback() is True
    assert wsgi_style.start_response() is True

    ## try canteen-style construction
    canteen_style, request, response, runtime = self._make_handler()
    assert hasattr(canteen_style, 'environ')
    assert hasattr(canteen_style, 'environment')
    assert canteen_style.environ['sample'] == 'hi'
    assert canteen_style.callback() is True
    assert canteen_style.start_response() is True
    assert canteen_style.request is request
    assert canteen_style.response is response
    assert canteen_style.runtime is runtime

  def test_template_context(self):

    """ Test `Handler.template_context` """

    handler, request, response, runtime = self._make_handler()
    context = handler.template_context

    # test top-level stuff
    for entry in (
      'handler', 'config', 'runtime', 'http', 'wsgi', 'cache',
      'asset', 'services', 'output', 'link', 'route'):
      assert entry in context, "%s not found in template context" % entry

    # base, HTTP and WSGI
    assert context['handler'] is handler
    assert context['runtime'] is runtime
    assert isinstance(context['config'], dict)
    assert context['wsgi']['callback']() is True
    assert context['http']['request'] is request
    assert context['wsgi']['callback']() is True
    assert context['http']['response'] is response
    assert context['wsgi']['start_response']() is True
    assert context['wsgi']['environ']['sample'] == 'hi'

    # links and routes
    assert callable(context['link'])
    assert callable(context['route']['build'])
    assert callable(context['route']['resolve'])

  def test_respond(self):

    """ Test `Handler.respond` interface """

    handler, request, response, runtime = self._make_handler(True)

    content = '<b>hi i am content</b>'
    handler.respond(content)

    assert handler.status is 200
    assert handler.response.response == content
    assert handler.response.status_code is 200

  def test_dispatch(self):

    """ Test `Handler` __call__ dispatch """

    content = '<b>hi sup</b>'


    class SubHandler(handler.Handler):

      """ I am an example handler """

      get_was_called = False

      def GET(self):

        """ I am an example GET method """

        self.__class__.get_was_called = True
        self.respond(content)

    _handler, request, response, runtime = self._make_handler(True, SubHandler)
    response = _handler({})

    assert SubHandler.get_was_called
    assert response.status_code is 200
    assert response.response == content

  def test_dispatch_direct(self):

    """ Test `Handler` direct __call__ dispatch """

    content = '<b>hi sup</b>'


    class SubHandler(handler.Handler):

      """ I am an example handler """

      get_was_called = False

      def GET(self):

        """ I am an example GET method """

        self.__class__.get_was_called = True
        self.respond(content)

    _handler, request, response, runtime = self._make_handler(True, SubHandler)
    response = _handler({}, direct=True)

    assert SubHandler.get_was_called
    assert response is _handler
    assert _handler.status is 200
    assert _handler.response.status_code is 200
    assert _handler.response.response == content

  def test_prepare_hook(self):

    """ Test that `Handler.prepare` is called before dispatch """

    content = '<b>hi sup</b>'


    class SubHandler(handler.Handler):

      """ I am an example handler """

      get_was_called = False
      prepare_tripped = False

      def prepare(self, url_args, direct):

        """ I prepare things """

        self.__class__.prepare_tripped = True

      def GET(self):

        """ I am an example GET method """

        if self.__class__.prepare_tripped:
          self.__class__.get_was_called = True
        self.respond(content)

    _handler, request, response, runtime = self._make_handler(True, SubHandler)
    response = _handler({})

    assert response.status_code is 200
    assert response.response == content
    assert SubHandler.prepare_tripped is True
    assert SubHandler.get_was_called is True

  def test_destroy_hook(self):

    """ Test that `Handler.destroy` is called after dispatch """

    content = '<b>hi sup</b>'


    class SubHandler(handler.Handler):

      """ I am an example handler """

      get_was_called = False
      prepare_tripped = False
      destroy_tripped = False

      def prepare(self, url_args, direct):

        """ I prepare things """

        self.__class__.prepare_tripped = True

      def GET(self):

        """ I am an example GET method """

        self.__class__.get_was_called = True
        self.respond(content)

      def destroy(self, _response):

        """ I destroy things """

        if self.__class__.get_was_called:
          self.__class__.destroy_tripped = True
        assert response, "response was lost between `GET` and `destroy`"

    _handler, request, response, runtime = self._make_handler(True, SubHandler)
    response = _handler({})

    assert response.status_code is 200
    assert response.response == content
    assert SubHandler.get_was_called is True
    assert SubHandler.prepare_tripped is True
    assert SubHandler.destroy_tripped is True

  def test_response_return(self):

    """ Test that `Handler.GET/POST/etc` can return a response """

    content = '<b>hi sup</b>'
    alt_content = '<b>goodbye_friend</b>'

    class SubHandler(handler.Handler):

      """ I am an example handler """

      def GET(self):

        """ I am an example GET method """

        self.respond(content)
        return alt_content  # cancels internal response

    _handler, request, response, runtime = self._make_handler(True, SubHandler)
    response = _handler({})

    assert response == alt_content

  def test_invalid_method(self):

    """ Test calling an invalid method on `Handler` (should raise HTTP405) """

    class SubHandler(handler.Handler):

      """ I am an example handler """

      def GET(self):

        """ I am an example GET method """

    _handler, request, response, runtime = self._make_handler(
      True,
      SubHandler,
      _environ={'REQUEST_METHOD': 'POST'})

    with self.assertRaises(MethodNotAllowed):
      _handler({})

  def test_full_render(self):

    """ Test the full template render flow from `Handler`'s perspective """

    h, rr, rs, rn = self._make_handler()

    # make working config
    c = Config()
    _cfg = c.blocks.get('app', {})
    _cfg['paths'] = _cfg.get('paths', {})
    _cfg['paths']['templates'] = _cfg['paths'].get('templates', {})
    _cfg['paths']['templates']['source'] = template._FRAMEWORK_TEMPLATE_SOURCES
    _cfg['paths']['templates']['compiled'] = 'canteen.templates.compiled'

    c.blocks['app'] = _cfg

    h.__dict__['config'] = c

    h.runtime.config = c
    result = h.render('base.html')
    assert result
